# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

# File created my Massimo Di Pierro

from itertools import combinations

try:
    from html import escape
except ImportError:
    # cgi is decprecated in python 3
    from cgi import escape
import json


def to_html(net, filename, respect_switches=True, include_lines=True, include_trafos=True, show_tables=True):
    """
    Saves a pandapower Network to an html file.

    INPUT:
        **net** (dict) - The pandapower format network

        **filename** (string) - The absolute or relative path to the input file.

    OPTIONAL:
        **respect_switches** (boolean, True) - True: open line switches are being considered (no edge between nodes)
                                               False: open line switches are being ignored

        **include_lines** (boolean, True) - determines, whether lines get converted to edges

        **include_trafos** (boolean, True) - determines, whether trafos get converted to edges

        **show_tables** (boolean, True) - shows pandapower element tables

    """
    if not filename.endswith(".html"):
        raise Exception("Please use .html to save pandapower networks!")
    with open(filename, "w") as f:
        html_str = _net_to_html(net, respect_switches, include_lines, include_trafos, show_tables)
        f.write(html_str)
        f.close()

class Raw(object):
    def __init__(self, html):
        self.html = html


class Tag(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, *args, **kwargs):
        attr = ' ' + ' '.join('%s="%s"' % (k, escape(v)) for k, v in kwargs.items())
        contents = ''.join(a.html if isinstance(a, Raw) else escape(str(a)) for a in args)
        return Raw('<%s%s>%s</%s>' % (self.name, attr.rstrip(), contents, self.name))

def _net_to_graphjs(net, respect_switches=True, include_lines=True, include_trafos=True, images=None):
    """
     Converts a pandapower network into a graph for vis.js
     which contains a simplified representation
     of a network's topology, reduced to nodes and edges. Busses are being represented by nodes
     (Note: only buses with in_service = 1 appear in the graph), edges represent physical
     connections between buses (typically lines or trafos).

     INPUT:
        **net** (pandapowerNet) - variable that contains a pandapower network


     OPTIONAL:
        **respect_switches** (boolean, True) - True: open line switches are being considered
                                                     (no edge between nodes)
                                               False: open line switches are being ignored

        **include_lines** (boolean, True) - determines, whether lines get converted to edges

        **include_trafos** (boolean, True) - determines, whether trafos get converted to edges

        **show_tables** (boolean, True) - shows pandapower element tables

     EXAMPLE:

         from pandapower.html import _net_to_graphjs
         print _net_to_graph(net, respect_switches = False)
    """
    nodes = [{'id': int(id), 'label': str(id)} for id in net.bus[net.bus.in_service == 1].index]
    edges = []

    if images:
        for node in nodes:
            node['shape'] = 'image'
            node['image'] = images.get(str(node['id']), images['default'])

    if include_lines:
        # lines with open switches can be excluded
        nogolines = set(net.switch.element[(net.switch.et == "l") & (net.switch.closed == 0)]) \
            if respect_switches else set()

        edges += [{'from': int(fb),
                   'to': int(tb),
                   'label': '%.2f kA %.2f Km' % (imax, length)}
                  for fb, tb, length, idx, inservice, imax in
                  list(zip(net.line.from_bus, net.line.to_bus, net.line.length_km,
                           net.line.index, net.line.in_service, net.line.max_i_ka))
                  if inservice == 1 and not idx in nogolines]

        edges += [{'from': int(fb), 'to': int(tb)}
                  for fb, tb, idx, inservice in
                  list(zip(net.impedance.from_bus, net.impedance.to_bus,
                           net.impedance.index, net.impedance.in_service))
                  if inservice == 1]

        edges += [{'from': int(fb),
                   'to': int(tb), 'dashes': 'true',
                   'label': '%.2f kA %.2f Km' % (imax, length)}
                  for fb, tb, length, idx, inservice, imax in
                  list(zip(net.line.from_bus, net.line.to_bus, net.line.length_km,
                           net.line.index, net.line.in_service, net.line.max_i_ka))
                  if inservice != 1 or idx in nogolines]

        edges += [{'from': int(fb), 'to': int(tb), 'dashes': 'true'}
                  for fb, tb, idx, inservice in
                  list(zip(net.impedance.from_bus, net.impedance.to_bus,
                           net.impedance.index, net.impedance.in_service))
                  if inservice != 1]

    if include_trafos:
        nogotrafos = set(net.switch.element[(net.switch.et == "t") & (net.switch.closed == 0)])
        edges += [{'from': int(hvb), 'to': int(lvb), 'label': 'trafo'}
                  for hvb, lvb, idx, inservice in
                  list(zip(net.trafo.hv_bus, net.trafo.lv_bus,
                           net.trafo.index, net.trafo.in_service))
                  if inservice == 1 and not idx in nogotrafos]
        for _, t3tab in net.trafo3w.iterrows():
            edges += [{'from': int(bus1),'to': int(bus2),'label': 'trafo3'}
                      for bus1, bus2 in combinations([t3tab.hv_bus, t3tab.mv_bus, t3tab.lv_bus], 2)
                      if t3tab.in_service]

    # add bus-bus switches
    bs = net.switch[(net.switch.et == "b") &
                    ((net.switch.closed == 1) | (not respect_switches))]
    edges += [{'from': int(b), 'to': int(e), 'label': 'switch'}
              for b, e, i in list(zip(bs.bus, bs.element, bs.index))]

    script = "var data = {nodes: new vis.DataSet(%s), edges: new vis.DataSet(%s)};" % (
        json.dumps(nodes), json.dumps(edges))
    script += "var container = document.getElementById('net');"
    script += "var network = new vis.Network(container, data);"
    script += "network.setOptions({interaction: {zoomView: true}});"

    return script


def _net_to_html(net, respect_switches=True, include_lines=True, include_trafos=True, show_tables=True, images=None):
    """
     Converts a pandapower network into an html page which contains a simplified representation
     of a network's topology, reduced to nodes and edges. Busses are being represented by nodes
     (Note: only buses with in_service = 1 appear in the graph), edges represent physical
     connections between buses (typically lines or trafos).

     INPUT:
        **net** (pandapowerNet) - variable that contains a pandapower network


     OPTIONAL:
        **respect_switches** (boolean, True) - True: open line switches are being considered
                                                     (no edge between nodes)
                                               False: open line switches are being ignored

        **include_lines** (boolean, True) - determines, whether lines get converted to edges

        **include_trafos** (boolean, True) - determines, whether trafos get converted to edges

        **show_tables** (boolean, True) - shows pandapower element tables

        **images** (dict) - an optional dict of images to map a node id into an image

     EXAMPLE:

         from pandapower.html import to_html
         html = to_html(net, respect_switches = False)
         open('C:\\index.html', "w").write(html)

    """

    HTML, HEAD, STYLE, BODY, DIV = Tag('html'), Tag('head'), Tag('style'), Tag('body'), Tag('div')
    TABLE, TR, TH, TD, SCRIPT = Tag('table'), Tag('tr'), Tag('th'), Tag('td'), Tag('script')
    H2 = Tag('h2')

    style = 'tr:first {background:#e1e1e1;} th,td {text-align:center; border:1px solid #e1e1e1;}'

    script = _net_to_graphjs(net, respect_switches, include_lines, include_trafos,  images)

    tables = []
    if show_tables:
        for name in ['bus', 'trafo', 'line', 'dcline', 'load', 'asymmetric_load', 'ext_grid', 'gen', 'sgen',
                     'res_bus', 'res_trafo', 'res_line', 'res_dcline', 'res_load', 'res_ext_grid', 'res_gen', 'res_sgen',
                     'res_bus_3ph', 'res_trafo_3ph', 'res_line_3ph', 'res_load_3ph', 'res_asymmetric_load_3ph', 'res_ext_grid_3ph', 'res_sgen_3ph']:
            item = getattr(net, name)
            if len(item):
                table = TABLE(TR(*map(TH, item.columns)),
                              *[TR(*map(TD, row)) for row in item.values])
                tables.append(DIV(H2(name), table))

    page = HTML(
        HEAD(STYLE(style)),
        BODY(DIV(*tables), DIV(id='net', style="border:1px solid #f1f1f1;max-width:90%")),
        SCRIPT(src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.18.1/vis.min.js"),
        SCRIPT(Raw(script))
    )

    return page.html



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

try:
    import pandaplan.core.pplog as pplog
except:
    import logging as pplog

from pandapower.io_utils import JSONSerializableClass

logger = pplog.getLogger(__name__)


class DataSource(JSONSerializableClass):
    """
    The DataSource class is a skeleton for data sources such as pandas DataFrames
    Controllers call get_time_step_values(time) in each time step to get values from the data source
    """

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__class__.__name__

    def get_time_step_value(self, time_step, profile_name, scale_factor=1.0):  # pragma: no cover
        """
        This method retrieves values of the data source according to the given parameters.
        For actual parameters look into the DataSource you are actually using.
        """
        raise NotImplementedError("Subclasses should implement this!")



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import inspect

from pandapower.auxiliary import _check_bus_index_and_print_warning_if_high, \
    _check_gen_index_and_print_warning_if_high, _init_runpp_options, _init_rundcopp_options, \
    _init_rundcpp_options, _init_runopp_options, _internal_stored
from pandapower.opf.validate_opf_input import _check_necessary_opf_parameters
from pandapower.powerflow import _powerflow, _recycled_powerflow
from pandapower.optimal_powerflow import _optimal_powerflow

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

try:
    from power_grid_model import PowerGridModel, CalculationType, CalculationMethod
    from power_grid_model.errors import PowerGridError
    from power_grid_model.validation import validate_input_data
    from power_grid_model_io.converters import PandaPowerConverter
    PGM_IMPORTED = True
except ImportError:
    PGM_IMPORTED = False

logger = logging.getLogger(__name__)


def set_user_pf_options(net, overwrite=False, **kwargs):
    """
    This function sets the 'user_pf_options' dict for net. These options overrule
    net.__internal_options once they are added to net. These options are used in configuration of
    load flow calculation.
    At the same time, user-defined arguments for pandapower.runpp() always have a higher priority.
    To remove user_pf_options, set overwrite=True and provide no additional arguments

    :param net: pandaPower network
    :param overwrite: specifies whether the user_pf_options is removed before setting new options
    :param kwargs: load flow options, e. g. tolerance_mva = 1e-3
    :return: None
    """
    standard_parameters = ['calculate_voltage_angles', 'trafo_model', 'check_connectivity', 'mode',
                           'copy_constraints_to_ppc', 'switch_rx_ratio', 'enforce_q_lims',
                           'recycle', 'voltage_depend_loads', 'consider_line_temperature', 'delta',
                           'trafo3w_losses', 'init', 'init_vm_pu', 'init_va_degree', 'init_results',
                           'tolerance_mva', 'trafo_loading', 'numba', 'ac', 'algorithm',
                           'max_iteration', 'v_debug', 'run_control', 'distributed_slack', 'lightsim2grid',
                           'tdpf', 'tdpf_delay_s', 'tdpf_update_r_theta']

    if overwrite or 'user_pf_options' not in net.keys():
        net['user_pf_options'] = dict()

    net.user_pf_options.update({key: val for key, val in kwargs.items()
                                if key in standard_parameters})

    additional_kwargs = {key: val for key, val in kwargs.items()
                         if key not in standard_parameters}

    # this part is to inform user and to make typos in parameters visible
    if len(additional_kwargs) > 0:
        logger.info('parameters %s are not in the list of standard options' % list(
            additional_kwargs.keys()))

        net.user_pf_options.update(additional_kwargs)


def runpp(net, algorithm='nr', calculate_voltage_angles=True, init="auto",
          max_iteration="auto", tolerance_mva=1e-8, trafo_model="t",
          trafo_loading="current", enforce_q_lims=False, check_connectivity=True,
          voltage_depend_loads=True, consider_line_temperature=False,
          run_control=False, distributed_slack=False, tdpf=False, tdpf_delay_s=None, **kwargs):
    """
    Runs a power flow

    INPUT:
        **net** - The pandapower format network

    OPTIONAL:
        **algorithm** (str, "nr") - algorithm that is used to solve the power flow problem.

            The following algorithms are available:

                - "nr" Newton-Raphson (pypower implementation with numba accelerations)
                - "iwamoto_nr" Newton-Raphson with Iwamoto multiplier (maybe slower than NR but more robust)
                - "bfsw" backward/forward sweep (specially suited for radial and weakly-meshed networks)
                - "gs" gauss-seidel (pypower implementation)
                - "fdbx" fast-decoupled (pypower implementation)
                - "fdxb" fast-decoupled (pypower implementation)

        **calculate_voltage_angles** (str or bool, True) - consider voltage angles in loadflow calculation

            If True, voltage angles of ext_grids and transformer shifts are considered in the
            loadflow calculation. Considering the voltage angles is only necessary in meshed
            networks that are usually found in higher voltage levels. calculate_voltage_angles
            in "auto" mode defaults to:

                - True, if the network voltage level is above 70 kV
                - False otherwise

            The network voltage level is defined as the maximum rated voltage of any bus in the network that
            is connected to a line.

        **init** (str, "auto") - initialization method of the loadflow
        pandapower supports four methods for initializing the loadflow:

            - "auto" - init defaults to "dc" if calculate_voltage_angles is True or "flat" otherwise
            - "flat"- flat start with voltage of 1.0pu and angle of 0° at all PQ-buses and 0° for PV buses as initial solution, the slack bus is initialized with the values provided in net["ext_grid"]
            - "dc" - initial DC loadflow before the AC loadflow. The results of the DC loadflow are used as initial solution for the AC loadflow. Note that the DC loadflow only calculates voltage angles at PQ and PV buses, voltage magnitudes are still flat started.
            - "results" - voltage vector of last loadflow from net.res_bus is used as initial solution. This can be useful to accelerate convergence in iterative loadflows like time series calculations.

        Considering the voltage angles might lead to non-convergence of the power flow in flat start.
        That is why in "auto" mode, init defaults to "dc" if calculate_voltage_angles is True or "flat" otherwise

        **max_iteration** (int, "auto") - maximum number of iterations carried out in the power flow algorithm.

            In "auto" mode, the default value depends on the power flow solver:

                - 10 for "nr"
                - 100 for "bfsw"
                - 1000 for "gs"
                - 30 for "fdbx"
                - 30 for "fdxb"
                - 30 for "nr" with "tdpf"

        **tolerance_mva** (float, 1e-8) - loadflow termination condition referring to P / Q mismatch of node power in MVA

        **trafo_model** (str, "t")  - transformer equivalent circuit model
        pandapower provides two equivalent circuit models for the transformer:

            - "t" - transformer is modeled as equivalent with the T-model.
            - "pi" - transformer is modeled as equivalent PI-model. This is not recommended, since it is less exact than the T-model. It is only recommended for valdiation with other software that uses the pi-model.

        **trafo_loading** (str, "current") - mode of calculation for transformer loading

            Transformer loading can be calculated relative to the rated current or the rated power. In both cases the overall transformer loading is defined as the maximum loading on the two sides of the transformer.

            - "current"- transformer loading is given as ratio of current flow and rated current of the transformer. This is the recommended setting, since thermal as well as magnetic effects in the transformer depend on the current.
            - "power" - transformer loading is given as ratio of apparent power flow to the rated apparent power of the transformer.

        **enforce_q_lims** (bool, False) - respect generator reactive power limits

            If True, the reactive power limits in net.gen.max_q_mvar/min_q_mvar are respected in the
            loadflow. This is done by running a second loadflow if reactive power limits are
            violated at any generator, so that the runtime for the loadflow will increase if reactive
            power has to be curtailed.

            Note: enforce_q_lims only works if algorithm="nr"!


        **check_connectivity** (bool, True) - Perform an extra connectivity test after the conversion from pandapower to PYPOWER

            If True, an extra connectivity test based on SciPy Compressed Sparse Graph Routines is perfomed.
            If check finds unsupplied buses, they are set out of service in the ppc

        **voltage_depend_loads** (bool, True) - consideration of voltage-dependent loads. If False, net.load.const_z_percent and net.load.const_i_percent are not considered, i.e. net.load.p_mw and net.load.q_mvar are considered as constant-power loads.

        **consider_line_temperature** (bool, False) - adjustment of line impedance based on provided
            line temperature. If True, net.line must contain a column "temperature_degree_celsius".
            The temperature dependency coefficient alpha must be provided in the net.line.alpha
            column, otherwise the default value of 0.004 is used

        **distributed_slack** (bool, False) - Distribute slack power
            according to contribution factor weights for external grids
            and generators.

        **tdpf** (bool, False) - Temperature Dependent Power Flow (TDPF). If True, line temperature is calculated based on the TDPF parameters in net.line table.

        **tdpf_delay_s** (float, None) - TDPF parameter, specifies the time delay in s to consider thermal inertia of conductors.


        **KWARGS**:

        **lightsim2grid** ((bool,str), "auto") - whether to use the package lightsim2grid for power
        flow backend. For more details on compatibility, check out pandapower's documentation.

        **numba** (bool, True) - Activation of numba JIT compiler in the newton solver.
        If set to True, the numba JIT compiler is used to generate matrices for the powerflow,
        which leads to significant speed improvements.

        **switch_rx_ratio** (float, 2) - rx_ratio of bus-bus-switches. If the impedance of switches
        defined in net.switch.z_ohm is zero, buses connected by a closed bus-bus switch are fused to
        model an ideal bus. Closed bus-bus switches, whose impedance z_ohm is not zero, are modelled
        as branches with resistance and reactance according to net.switch.z_ohm and switch_rx_ratio.

        **delta_q** - Reactive power tolerance for option "enforce_q_lims" in kvar - helps convergence in some cases.

        **trafo3w_losses** - defines where open loop losses of three-winding transformers are considered. Valid options are "hv", "mv", "lv" for HV/MV/LV side or "star" for the star point.

        **v_debug** (bool, False) - if True, voltage values in each newton-raphson iteration are logged in the ppc

        **init_vm_pu** (string/float/array/Series, None) - Allows to define initialization specifically for voltage magnitudes. Only works with init == "auto"!

            - "auto": all buses are initialized with the mean value of all voltage controlled elements in the grid
            - "flat" for flat start from 1.0
            - "results": voltage magnitude vector is taken from result table
            - a float with which all voltage magnitudes are initialized
            - an iterable with a voltage magnitude value for each bus (length and order has to match with the buses in net.bus)
            - a pandas Series with a voltage magnitude value for each bus (indexes have to match the indexes in net.bus)

        **init_va_degree** (string/float/array/Series, None) - Allows to define initialization specifically for voltage angles. Only works with init == "auto"!

            - "auto": voltage angles are initialized from DC power flow if angles are calculated or as 0 otherwise
            - "dc": voltage angles are initialized from DC power flow
            - "flat" for flat start from 0
            - "results": voltage angle vector is taken from result table
            - a float with which all voltage angles are initialized
            - an iterable with a voltage angle value for each bus (length and order has to match with the buses in net.bus)
            - a pandas Series with a voltage angle value for each bus (indexes have to match the indexes in net.bus)

        **recycle** (dict, none) - Reuse of internal powerflow variables for time series calculation

            Contains a dict with the following parameters:
            bus_pq: If True PQ values of buses are updated
            trafo: If True trafo relevant variables, e.g., the Ybus matrix, is recalculated
            gen: If True Sbus and the gen table in the ppc are recalculated

        **neglect_open_switch_branches** (bool, False) - If True no auxiliary buses are created for branches when switches are opened at the branch. Instead branches are set out of service

        **tdpf_update_r_theta** (bool, True) - TDPF parameter, whether to update R_Theta in Newton-Raphson or to assume a constant R_Theta (either from net.line.r_theta, if set, or from a calculation based on the thermal model of Ngoko et.al.)
    """

    # if dict 'user_pf_options' is present in net, these options overrule the net._options
    # except for parameters that are passed by user
    if isinstance(kwargs.get("recycle", None), dict) and _internal_stored(net):
        _recycled_powerflow(net, **kwargs)
        return

    if run_control and net.controller.in_service.any():
        from pandapower.control import run_control
        parameters = {**locals(), **kwargs}
        # disable run control for inner loop to avoid infinite loop
        parameters["run_control"] = False
        run_control(**parameters)
    else:
        passed_parameters = _passed_runpp_parameters(locals())
        _init_runpp_options(net, algorithm=algorithm,
                            calculate_voltage_angles=calculate_voltage_angles,
                            init=init, max_iteration=max_iteration, tolerance_mva=tolerance_mva,
                            trafo_model=trafo_model, trafo_loading=trafo_loading,
                            enforce_q_lims=enforce_q_lims, check_connectivity=check_connectivity,
                            voltage_depend_loads=voltage_depend_loads,
                            consider_line_temperature=consider_line_temperature,
                            tdpf=tdpf, tdpf_delay_s=tdpf_delay_s,
                            distributed_slack=distributed_slack,
                            passed_parameters=passed_parameters, **kwargs)
        _check_bus_index_and_print_warning_if_high(net)
        _check_gen_index_and_print_warning_if_high(net)
        _powerflow(net, **kwargs)


def runpp_pgm(net, algorithm="nr", max_iterations=20, error_tolerance_vm_pu=1e-8, symmetric=True, validate_input=False):
    """
        Runs powerflow using power-grid-model library

        INPUT:
            **net** - The pandapower format network

        OPTIONAL:
            **symmetric** (bool, True) -

            - True: three-phase symmetric calculation, even for asymmetric loads/generations
            - False: three-phase asymmetric calculation

            **algorithm** (str, "nr") - Algorithms available in power-grid-model.
            Check power-grid-model documentation for detailed information on the algorithms.

            - "nr" - Newton Raphson algorithm
            - "bfsw" - Iterative current algorithm. Similar to backward-forward sweep algorithm
            - "lc" - Linear current approximation algorithm
            - "lin" - Linear approximation algorithm

            **error_tolerance_u_pu** (float, 1e-8) - error tolerance for voltage in p.u.

            **max_iterations** (int, 20) - Maximum number of iterations for algorithms.
            No effect on linear approximation algorithms.

            **validate_input** (bool, False) - Validate input data to be used for power-flow in power-grid-model.
            It is recommended to use pandapower.diagnostic tool prior.
    """
    if not PGM_IMPORTED:
        raise ImportError(
            "Power Grid Model import failed. Try using `pip install pandapower[pgm]` to install the required packages."
        )

    # 1. Determine the Power Grid Model calculation type corresponding to the algorithm:
    algorithm_map = {
        "nr": CalculationMethod.newton_raphson,
        "lin": CalculationMethod.linear,
        "bfsw": CalculationMethod.iterative_current,
        "lc": CalculationMethod.linear_current
    }
    if algorithm not in algorithm_map:
        raise KeyError(f"Invalid algorithm '{algorithm}'")
    calculation_method = algorithm_map[algorithm]

    # 2. Convert the pandapower data to the power grid model format. Ignoring the 'extra info', which is a
    #    dictionary representation of the internal state of the PandaPowerConverter.
    pgm_converter = PandaPowerConverter()
    pgm_input_data, _extra_info = pgm_converter.load_input_data(net, make_extra_info=False)

    # 3. If required, validate the input data
    if validate_input:
        validation_errors = validate_input_data(
            pgm_input_data, calculation_type=CalculationType.power_flow, symmetric=symmetric
        )

        # If there were validation errors, convert the errors to a human readable test and log each error for
        # each element in the input data.
        if validation_errors:
            for i, error in enumerate(validation_errors, start=1):
                pp_obj = (pgm_converter.lookup_id(pgm_id) for pgm_id in error.ids)
                pp_obj = (f"{obj['table']}-{obj['index']}" for obj in pp_obj)
                pp_obj = ", ".join(pp_obj)
                logger.error(f"{i}. Power Grid Model validation error: Check {pp_obj}")
                logger.debug(f"{i}. Native Power Grid Model error: {error}")

    # 4. Create a PowerGridModel and calculate the powerflow
    try:
        pgm = PowerGridModel(input_data=pgm_input_data)
        output_data = pgm.calculate_power_flow(
            symmetric=symmetric,
            error_tolerance=error_tolerance_vm_pu,
            max_iterations=max_iterations,
            calculation_method=calculation_method
        )
        net["converged"] = True

    # Stop execution if a power PowerGridError was raised and log the error
    except PowerGridError as ex:
        logger.critical(f"Internal {type(ex).__name__} occurred!")
        logger.debug(str(ex))
        if not validate_input:
            logger.info("Use validate_input=True to validate your input data.")
        net["converged"] = False
        return

    # 5. Convert the Power Grid Model output data back to the pandapower format and update the pandapower net
    converted_output_data = pgm_converter.convert(data=output_data)
    for table in converted_output_data.keys():
        net[table] = converted_output_data[table]


def rundcpp(net, trafo_model="t", trafo_loading="current", recycle=None, check_connectivity=True,
            switch_rx_ratio=2, trafo3w_losses="hv", **kwargs):
    """
    Runs PANDAPOWER DC Flow

    INPUT:
        **net** - The pandapower format network

    OPTIONAL:
        **trafo_model** (str, "t")  - transformer equivalent circuit model
        pandapower provides two equivalent circuit models for the transformer:

        - "t" - transformer is modeled as equivalent with the T-model. This is consistent with PowerFactory and is also more accurate than the PI-model. We recommend using this transformer model.
        - "pi" - transformer is modeled as equivalent PI-model. This is consistent with Sincal, but the method is questionable since the transformer is physically T-shaped. We therefore recommend the use of the T-model.

        **trafo_loading** (str, "current") - mode of calculation for transformer loading

        Transformer loading can be calculated relative to the rated current or the rated power. In both cases the overall transformer loading is defined as the maximum loading on the two sides of the transformer.

        - "current"- transformer loading is given as ratio of current flow and rated current of the transformer. This is the recommended setting, since thermal as well as magnetic effects in the transformer depend on the current.
        - "power" - transformer loading is given as ratio of apparent power flow to the rated apparent power of the transformer.

        **check_connectivity** (bool, False) - Perform an extra connectivity test after the conversion from pandapower to PYPOWER

        If true, an extra connectivity test based on SciPy Compressed Sparse Graph Routines is perfomed.
        If check finds unsupplied buses, they are put out of service in the PYPOWER matrix

        **switch_rx_ratio** (float, 2) - rx_ratio of bus-bus-switches. If the impedance of switches
        defined in net.switch.z_ohm is zero, buses connected by a closed bus-bus switch are fused to
        model an ideal bus. Closed bus-bus switches, whose impedance z_ohm is not zero, are modelled
        as branches with resistance and reactance according to net.switch.z_ohm and switch_rx_ratio.

        **trafo3w_losses** (str, "hv") - defines where open loop losses of three-winding transformers are considered. Valid options are "hv", "mv", "lv" for HV/MV/LV side or "star" for the star point.

        **kwargs** - options to use for PYPOWER.runpf
    """
    _init_rundcpp_options(net, trafo_model=trafo_model, trafo_loading=trafo_loading,
                          recycle=recycle, check_connectivity=check_connectivity,
                          switch_rx_ratio=switch_rx_ratio, trafo3w_losses=trafo3w_losses, **kwargs)

    if isinstance(recycle, dict) and _internal_stored(net, ac=False):
        _recycled_powerflow(net, recycle=recycle, **kwargs)
        return

    _check_bus_index_and_print_warning_if_high(net)
    _check_gen_index_and_print_warning_if_high(net)
    _powerflow(net, **kwargs)


def runopp(net, verbose=False, calculate_voltage_angles=True, check_connectivity=True,
           suppress_warnings=True, switch_rx_ratio=2, delta=1e-10, init="flat", numba=True,
           trafo3w_losses="hv", consider_line_temperature=False, **kwargs):
    """
        Runs the  pandapower Optimal Power Flow.
        Flexibilities, constraints and cost parameters are defined in the pandapower element tables.

        Flexibilities can be defined in net.sgen / net.gen /net.load / net.storage /net.ext_grid
        net.sgen.controllable if a static generator is controllable. If False,
        the active and reactive power are assigned as in a normal power flow. If True, the following
        flexibilities apply:

        - net.gen.min_p_mw / net.gen.max_p_mw
        - net.gen.min_q_mvar / net.gen.max_q_mvar
        - net.sgen.min_p_mw / net.sgen.max_p_mw
        - net.sgen.min_q_mvar / net.sgen.max_q_mvar
        - net.dcline.max_p_mw
        - net.dcline.min_q_to_mvar / net.dcline.max_q_to_mvar / net.dcline.min_q_from_mvar / net.dcline.max_q_from_mvar
        - net.ext_grid.min_p_mw / net.ext_grid.max_p_mw
        - net.ext_grid.min_q_mvar / net.ext_grid.max_q_mvar
        - net.load.min_p_mw / net.load.max_p_mw
        - net.load.min_q_mvar / net.load.max_q_mvar
        - net.storage.min_p_mw / net.storage.max_p_mw
        - net.storage.min_q_mvar / net.storage.max_q_mvar

        Controllable loads behave just like controllable static generators. It must be stated if they are controllable.
        Otherwise, they are not respected as flexibilities.
        Dc lines are controllable per default

        Network constraints can be defined for buses, lines and transformers the elements in the following columns:

        - net.bus.min_vm_pu / net.bus.max_vm_pu
        - net.line.max_loading_percent
        - net.trafo.max_loading_percent
        - net.trafo3w.max_loading_percent

        If the external grid ist controllable, the voltage setpoint of the external grid can be optimized within the
        voltage constraints by the OPF. The same applies to the voltage setpoints of the controllable generator elements.

        How these costs are combined into a cost function depends on the cost_function parameter.

        INPUT:
            **net** - The pandapower format network

        OPTIONAL:
            **verbose** (bool, False) - If True, some basic information is printed

            **suppress_warnings** (bool, True) - suppress warnings in pypower

                If set to True, warnings are disabled during the loadflow. Because of the way data is
                processed in pypower, ComplexWarnings are raised during the loadflow.
                These warnings are suppressed by this option, however keep in mind all other pypower
                warnings are suppressed, too.

            **init** (str, "flat") - init of starting opf vector. Options are "flat", "pf" or "results"

                Starting solution vector (x0) for opf calculations is determined by this flag. Options are:
                "flat" (default): starting vector is (upper bound - lower bound) / 2
                "pf": a power flow is executed prior to the opf and the pf solution is the starting vector. This may improve
                convergence, but takes a longer runtime (which are probably neglectible for opf calculations)
                "results": voltage magnitude vector is taken from result table

            **delta** (float, 1e-10) - power tolerance

            **trafo3w_losses** (str, "hv") - defines where open loop losses of three-winding transformers are considered. Valid options are "hv", "mv", "lv" for HV/MV/LV side or "star" for the star point.

            **consider_line_temperature** (bool, False) - adjustment of line impedance based on provided\
                line temperature. If True, net.line must contain a column "temperature_degree_celsius".\
                The temperature dependency coefficient alpha must be provided in the net.line.alpha\
                column, otherwise the default value of 0.004 is used

            **kwargs** - Pypower / Matpower keyword arguments:

            - OPF_VIOLATION (5e-6) constraint violation tolerance
            - PDIPM_COSTTOL (1e-6) optimality tolerance
            - PDIPM_GRADTOL (1e-6) gradient tolerance
            - PDIPM_COMPTOL (1e-6) complementarity condition (inequality) tolerance
            - PDIPM_FEASTOL (set to OPF_VIOLATION if not specified) feasibiliy (equality) tolerance
            - PDIPM_MAX_IT  (150) maximum number of iterations
            - SCPDIPM_RED_IT(20) maximum number of step size reductions per iteration
    """
    _check_necessary_opf_parameters(net, logger)
    _init_runopp_options(net, calculate_voltage_angles=calculate_voltage_angles,
                         check_connectivity=check_connectivity,
                         switch_rx_ratio=switch_rx_ratio, delta=delta, init=init, numba=numba,
                         trafo3w_losses=trafo3w_losses,
                         consider_line_temperature=consider_line_temperature, **kwargs)
    _check_bus_index_and_print_warning_if_high(net)
    _check_gen_index_and_print_warning_if_high(net)
    _optimal_powerflow(net, verbose, suppress_warnings, **kwargs)


def rundcopp(net, verbose=False, check_connectivity=True, suppress_warnings=True,
             switch_rx_ratio=0.5, delta=1e-10, trafo3w_losses="hv", **kwargs):
    """
    Runs the  pandapower Optimal Power Flow.
    Flexibilities, constraints and cost parameters are defined in the pandapower element tables.

    Flexibilities for generators can be defined in net.sgen / net.gen.
    net.sgen.controllable / net.gen.controllable signals if a generator is controllable. If False,
    the active and reactive power are assigned as in a normal power flow. If yes, the following
    flexibilities apply:
    - net.sgen.min_p_mw / net.sgen.max_p_mw
    - net.gen.min_p_mw / net.gen.max_p_mw
    - net.load.min_p_mw / net.load.max_p_mw

    Network constraints can be defined for buses, lines and transformers the elements in the following columns:
    - net.line.max_loading_percent
    - net.trafo.max_loading_percent
    - net.trafo3w.max_loading_percent

    INPUT:
        **net** - The pandapower format network

    OPTIONAL:
        **verbose** (bool, False) - If True, some basic information is printed

        **suppress_warnings** (bool, True) - suppress warnings in pypower

            If set to True, warnings are disabled during the loadflow. Because of the way data is
            processed in pypower, ComplexWarnings are raised during the loadflow.
            These warnings are suppressed by this option, however keep in mind all other pypower
            warnings are suppressed, too.

        **delta** (float, 1e-10) - power tolerance

        **trafo3w_losses** (str, "hv") - defines where open loop losses of three-winding transformers are considered. Valid options are "hv", "mv", "lv" for HV/MV/LV side or "star" for the star point.
    """
    if (not net.sgen.empty) & ("controllable" not in net.sgen.columns):
        logger.warning('Warning: Please specify sgen["controllable"]\n')

    if (not net.load.empty) & ("controllable" not in net.load.columns):
        logger.warning('Warning: Please specify load["controllable"]\n')

    _init_rundcopp_options(net, check_connectivity=check_connectivity,
                           switch_rx_ratio=switch_rx_ratio, delta=delta,
                           trafo3w_losses=trafo3w_losses, **kwargs)
    _check_bus_index_and_print_warning_if_high(net)
    _check_gen_index_and_print_warning_if_high(net)
    _optimal_powerflow(net, verbose, suppress_warnings, **kwargs)


def _passed_runpp_parameters(local_parameters):
    """
    Internal function to distinguish arguments for pandapower.runpp() that are explicitly passed by
    the user.
    :param local_parameters: locals() in the runpp() function
    :return: dictionary of explicitly passed parameters
    """
    net = local_parameters.pop("net")
    if "user_pf_options" not in net.keys() or len(net.user_pf_options) == 0:
        return None
    # default_parameters contains the parameters that are specified for the runpp function by default in its definition
    args, varargs, keywords, defaults, *_ = inspect.getfullargspec(runpp)
    default_parameters = dict(zip(args[1:], defaults))

    # we want to also include the parameters that are optional (passed in "kwargs")!
    # that is why we include the parameters that are also not in default_parameters
    # maybe the part "if key not in default_parameters.keys()" is redundant, idk
    kwargs_parameters = local_parameters.pop('kwargs', None)
    passed_parameters = {
        key: val for key, val in local_parameters.items()
        if key not in default_parameters.keys() or val != default_parameters.get(key, None)}
    # passed_parameters should have the "kwargs" parameters in the same level as other parameters
    # (not nested in "kwargs: {...}") for them to be considered later on, otherwise they will be ignored
    passed_parameters.update(kwargs_parameters)

    return passed_parameters



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import pandas as pd

from pandapower.plotting.generic_geodata import create_generic_coordinates
from pandapower.plotting.plotly.traces import create_bus_trace, create_line_trace, \
    create_trafo_trace, draw_traces
from pandapower.plotting.plotly.get_colors import get_plotly_color_palette
from pandapower.plotting.plotly.mapbox_plot import geo_data_to_latlong
from pandapower.topology import create_nxgraph, connected_components

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


def vlevel_plotly(net, respect_switches=True, use_line_geo=None, colors_dict=None, on_map=False,
                  projection=None, map_style='basic', figsize=1, aspectratio='auto', line_width=2,
                  bus_size=10, filename="temp-plot.html", auto_open=True):
    """
    Plots a pandapower network in plotly
    using lines/buses colors according to the voltage level they belong to.
    If no geodata is available, artificial geodata is generated. For advanced plotting see the
    tutorial

    INPUT:
        **net** - The pandapower format network.

    OPTIONAL:
        **respect_switches** (bool, True) - Respect switches when artificial geodata is created

        **use_line_geo** (bool, True) - defines if lines patches are based on net.line.geo
        of the lines (True) or on net.bus.geo of the connected buses (False)

        *colors_dict** (dict, None) - dictionary for customization of colors for each voltage level
        in the form: voltage : color

        **on_map** (bool, False) - enables using mapbox plot in plotly If provided geodata are not
        real geo-coordinates in lon/lat form, on_map will be set to False.

        **projection** (String, None) - defines a projection from which network geo-data will be
        transformed to lat-long. For each projection a string can be found at
        http://spatialreference.org/ref/epsg/

        **map_style** (str, 'basic') - enables using mapbox plot in plotly

            - 'streets'
            - 'bright'
            - 'light'
            - 'dark'
            - 'satellite'

        **figsize** (float, 1) - aspectratio is multiplied by it in order to get final image size

        **aspectratio** (tuple, 'auto') - when 'auto' it preserves original aspect ratio of the
        network geodata any custom aspectration can be given as a tuple, e.g. (1.2, 1)

        **line_width** (float, 1.0) - width of lines

        **bus_size** (float, 10.0) -  size of buses to plot.

        **filename** (str, "temp-plot.html") - filename / path to plot to. Should end on `*.html`

        **auto_open** (bool, True) - automatically open plot in browser

    OUTPUT:
        **figure** (graph_objs._figure.Figure) figure object

    """
    # getting connected componenets without consideration of trafos
    graph = create_nxgraph(net, include_trafos=False)
    vlev_buses = connected_components(graph)
    # getting unique sets of buses for each voltage level
    vlev_bus_dict = {}
    for vl_buses in vlev_buses:
        if net.bus.loc[list(vl_buses), 'vn_kv'].unique().shape[0] > 1:
            logger.warning('buses from the same voltage level does not have the same vn_kv !?')
        vn_kv = net.bus.loc[list(vl_buses), 'vn_kv'].unique()[0]
        if vlev_bus_dict.get(vn_kv):
            vlev_bus_dict[vn_kv].update(vl_buses)
        else:
            vlev_bus_dict[vn_kv] = vl_buses

    # create a default colormap for voltage levels
    nvlevs = len(vlev_bus_dict)
    colors = get_plotly_color_palette(nvlevs)
    colors_dict = colors_dict or dict(zip(vlev_bus_dict.keys(), colors))
    bus_groups = [(buses, colors_dict[vlev], f"{vlev} kV") for vlev, buses in vlev_bus_dict.items()]

    return _draw_colored_bus_groups_plotly(
        net, bus_groups, respect_switches=respect_switches,
        use_line_geo=use_line_geo, on_map=on_map, projection=projection,
        map_style=map_style, figsize=figsize, aspectratio=aspectratio, line_width=line_width,
        bus_size=bus_size, filename=filename, auto_open=auto_open)


def _draw_colored_bus_groups_plotly(
    net, bus_groups, respect_switches=True, use_line_geo=None,
    on_map=False, projection=None, map_style='basic', figsize=1, aspectratio='auto', line_width=2,
    bus_size=10, filename="temp-plot.html", auto_open=True):
    """
    Internal function of vlevel_plotly()

    **bus_groups** - list of tuples consisting of set of bus indices, color, legendgroup
    """
    # create geocoord if none are available
    if any(net.line.geo.isna()) and any(net.bus.geo.isna()):
        logger.warning("No or insufficient geodata available --> Creating artificial coordinates." +
                       " This may take some time")
        create_generic_coordinates(net, respect_switches=respect_switches)
        if on_map:
            logger.warning(
                "Map plots not available with artificial coordinates and will be disabled!")
            on_map = False

    # check if geodata are real geographycal lat/lon coordinates using geopy
    if on_map and projection is not None:
        geo_data_to_latlong(net, projection=projection)

    # if bus geodata is available, but no line geodata
    if use_line_geo is None:
        use_line_geo = False if any(net.line.geo.isna()) else True
    elif use_line_geo and any(net.line.geo.isna()):
        logger.warning(
            "No or insufficient line geodata available --> only bus geodata will be used.")
        use_line_geo = False

    # creating traces for buses and lines for each voltage level
    bus_traces = []
    line_traces = []
    traced_lines = set()
    for bus_group_legend in bus_groups:
        buses_vl, vlev_color, legend_group = bus_group_legend

        bus_trace_vlev = create_bus_trace(
            net, buses=buses_vl, size=bus_size, legendgroup=legend_group, color=vlev_color,
            trace_name=f'buses {legend_group}')
        if bus_trace_vlev is not None:
            bus_traces += bus_trace_vlev

        vlev_lines = net.line[net.line.from_bus.isin(buses_vl) &
                              net.line.to_bus.isin(buses_vl)].index.tolist()
        traced_lines |= set(vlev_lines)
        line_trace_vlev = create_line_trace(
            net, lines=vlev_lines, use_line_geo=use_line_geo,
            respect_switches=respect_switches, legendgroup=legend_group, color=vlev_color,
            width=line_width, trace_name=f'lines {legend_group}')
        if line_trace_vlev is not None:
            line_traces += line_trace_vlev

    # creating traces for other lines
    line_trace_other = create_line_trace(
        net, lines=net.line.index.difference(traced_lines).tolist(),
        use_line_geo=use_line_geo, respect_switches=respect_switches,
        color="grey", width=line_width)
    if line_trace_vlev is not None:
        line_traces += line_trace_other

    trafo_traces = create_trafo_trace(net, color='gray', width=line_width * 2)

    return draw_traces(line_traces + trafo_traces + bus_traces, showlegend=True,
                       aspectratio=aspectratio, on_map=on_map, map_style=map_style, figsize=figsize,
                       filename=filename, auto_open=auto_open)


if __name__ == '__main__':
    from pandapower.plotting.plotly import simple_plotly
    from pandapower.networks import mv_oberrhein
    from pandapower import runpp
    net = mv_oberrhein()
    vlevel_plotly(net)
    runpp(net)
    line_width = 2
    bus_size = 10
    use_line_geodata = None
    graph = create_nxgraph(net, include_trafos=False)
    vlev_buses = connected_components(graph)
    respect_switches = True

    # getting unique sets of buses for each voltage level
    vlev_bus_dict = {}
    for vl_buses in vlev_buses:
        if net.bus.loc[vl_buses, 'vn_kv'].unique().shape[0] > 1:
            logger.warning('buses from the same voltage level does not have the same vn_kv !?')
        vn_kv = net.bus.loc[vl_buses, 'vn_kv'].unique()[0]
        if vlev_bus_dict.get(vn_kv):
            vlev_bus_dict[vn_kv].update(vl_buses)
        else:
            vlev_bus_dict[vn_kv] = vl_buses

    # create a default colormap for voltage levels
    nvlevs = len(vlev_bus_dict)
    colors = get_plotly_color_palette(nvlevs)
    colors_dict = dict(zip(vlev_bus_dict.keys(), colors))

    # creating traces for buses and lines for each voltage level
    bus_traces = []
    line_traces = []
    for vn_kv, buses_vl in vlev_bus_dict.items():

        vlev_color = colors_dict[vn_kv]
        bus_trace_vlev = create_bus_trace(
            net, buses=buses_vl, size=bus_size, legendgroup=str(vn_kv),
            color=vlev_color, trace_name='buses {0} kV'.format(vn_kv))
        if bus_trace_vlev is not None:
            bus_traces += bus_trace_vlev

        vlev_lines = net.line[net.line.from_bus.isin(buses_vl) &
                              net.line.to_bus.isin(buses_vl)].index.tolist()
        print(vlev_lines)
        line_trace_vlev = create_line_trace(
            net, lines=vlev_lines, use_line_geodata=use_line_geodata,
            respect_switches=respect_switches, legendgroup=str(vn_kv), color="r",
            width=line_width, trace_name='lines {0} kV'.format(vn_kv))
        if line_trace_vlev is not None:
            line_traces += line_trace_vlev



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.auxiliary import read_from_net, write_to_net
from pandapower.control.controller.trafo_control import TrafoController


class ContinuousTapControl(TrafoController):
    """
    Trafo Controller with local tap changer voltage control.

    INPUT:
        **net** (attrdict) - Pandapower struct

        **element_index** (int) - ID of the trafo that is controlled

        **vm_set_pu** (float) - Maximum OLTC target voltage at bus in pu

    OPTIONAL:

        **tol** (float, 0.001) - Voltage tolerance band at bus in percent (default: 1% = 0.01pu)

        **side** (string, "lv") - Side of the transformer where the voltage is controlled

        **element** (string, "trafo") - Trafo type ("trafo" or "trafo3w")

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **check_tap_bounds** (bool, True) - In case of true the tap_bounds will be considered

        **drop_same_existing_ctrl** (bool, False) - Indicates if already existing controllers of the same type and with the same matching parameters (e.g. at same element) should be dropped
    """

    def __init__(self, net, element_index, vm_set_pu, tol=1e-3, side="lv", element="trafo", in_service=True,
                 check_tap_bounds=True, level=0, order=0, drop_same_existing_ctrl=False,
                 matching_params=None, **kwargs):
        if matching_params is None:
            matching_params = {"element_index": element_index, 'element': element}
        super().__init__(net, element_index=element_index, side=side, tol=tol, in_service=in_service,
                         element=element, level=level, order=order,
                         drop_same_existing_ctrl=drop_same_existing_ctrl,
                         matching_params=matching_params, **kwargs)

        self._set_t_nom(net)
        self.check_tap_bounds = check_tap_bounds
        self.vm_set_pu = vm_set_pu

    def _set_t_nom(self, net):
        vn_hv_kv = read_from_net(net, self.element, self.element_index, 'vn_hv_kv', self._read_write_flag)
        hv_bus = read_from_net(net, self.element, self.element_index, 'hv_bus', self._read_write_flag)
        vn_hv_bus_kv = read_from_net(net, "bus", hv_bus, 'vn_kv', self._read_write_flag)

        if self.element == "trafo3w" and self.side == "mv":
            vn_mv_kv = read_from_net(net, self.element, self.element_index, 'vn_mv_kv', self._read_write_flag)
            mv_bus = read_from_net(net, self.element, self.element_index, 'mv_bus', self._read_write_flag)
            vn_mv_bus_kv = read_from_net(net, "bus", mv_bus, 'vn_kv', self._read_write_flag)
            self.t_nom = vn_mv_kv / vn_hv_kv * vn_hv_bus_kv / vn_mv_bus_kv
        else:
            vn_lv_kv = read_from_net(net, self.element, self.element_index, 'vn_lv_kv', self._read_write_flag)
            lv_bus = read_from_net(net, self.element, self.element_index, 'lv_bus', self._read_write_flag)
            vn_lv_bus_kv = read_from_net(net, "bus", lv_bus, 'vn_kv', self._read_write_flag)
            self.t_nom = vn_lv_kv / vn_hv_kv * vn_hv_bus_kv / vn_lv_bus_kv

    def initialize_control(self, net):
        super().initialize_control(net)
        if not self.nothing_to_do(net):
            self._set_t_nom(net)  # in case some of the trafo elements change their in_service in between runs

    def control_step(self, net):
        """
        Implements one step of the ContinuousTapControl
        """
        if self.nothing_to_do(net):
            return

        delta_vm_pu = read_from_net(net, "res_bus", self.trafobus, 'vm_pu', self._read_write_flag) - self.vm_set_pu
        tc = delta_vm_pu / self.tap_step_percent * 100 / self.t_nom
        self.tap_pos = self.tap_pos + tc * self.tap_side_coeff * self.tap_sign
        if self.check_tap_bounds:
            self.tap_pos = np.clip(self.tap_pos, self.tap_min, self.tap_max)

        # WRITE TO NET
        # necessary in case the dtype of the column is int
        if net[self.element].tap_pos.dtype != "float":
            net[self.element].tap_pos = net[self.element].tap_pos.astype(float)
        write_to_net(net, self.element, self.element_index, "tap_pos", self.tap_pos, self._read_write_flag)

    def is_converged(self, net):
        """
        The ContinuousTapControl is converged, when the difference of the voltage between control steps is smaller
        than the Tolerance (tol).
        """
        if self.nothing_to_do(net):
            return True

        vm_pu = read_from_net(net, "res_bus", self.trafobus, "vm_pu", self._read_write_flag)
        # this is possible in case the trafo is set out of service by the connectivity check
        is_nan =  np.isnan(vm_pu)
        self.tap_pos = read_from_net(net, self.element, self.element_index, "tap_pos", self._read_write_flag)
        difference = 1 - self.vm_set_pu / vm_pu

        if self.check_tap_bounds:
            reached_limit = np.where(self.tap_side_coeff * self.tap_sign == 1,
                                     (vm_pu < self.vm_set_pu) & (self.tap_pos == self.tap_min) |
                                     (vm_pu > self.vm_set_pu) & (self.tap_pos == self.tap_max),
                                     (vm_pu < self.vm_set_pu) & (self.tap_pos == self.tap_max) |
                                     (vm_pu > self.vm_set_pu) & (self.tap_pos == self.tap_min))
            converged = np.logical_or(reached_limit, np.abs(difference) < self.tol)
        else:
            converged = np.abs(difference) < self.tol

        return np.all(np.logical_or(converged, is_nan))



# -*- coding: utf-8 -*-
"""
This file included various test network for modelling and simulation different pandapower modules

"""
import pandas as pd

from pandapower.create import create_empty_network, create_buses, create_ext_grid, create_lines, create_switches, \
    create_loads

pd.Series(dtype='float64')


def three_radial_bus_net():
    net = create_empty_network()  # create an empty network
    create_buses(net, nr_buses=16, vn_kv=20, index=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15], name=None,
                 type="n",
                 geodata=[(0, 0), (0, -1), (0, -2), (-2, -3), (-2, -4), (-2, -5), (0, -6),
                          (0, -4.5), (0, -5), (0, -6), (0, -7), (0, -8), (2, -8), (2, -9), (2, -10), (2, -12)])

    # external grids
    create_ext_grid(net, 0, vm_pu=1.0, va_degree=0, s_sc_max_mva=100, s_sc_min_mva=50, rx_max=0.1, rx_min=0.1)

    create_lines(net, from_buses=[0, 1, 2, 3, 4, 5, 2, 7, 8, 9, 10, 9, 12, 13, 14],
                 to_buses=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
                 length_km=[5, 4, 4, 0.5, 0.5, 1, 2, 4, 4, 5, 6, 5, 3, 2, 1], std_type="NAYY 4x50 SE",
                 name=None, index=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14], df=1., parallel=1)

    net.line["endtemp_degree"] = 250

    create_switches(net, buses=[0, 1, 2, 3, 4, 5, 2, 7, 8, 9, 10, 9, 12, 13, 14], elements=
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14], et='l', type="CB_IDTOC")

    # Define load
    create_loads(net, buses=[5, 11, 15], p_mw=[5, 5, 4], q_mvar=[1, 1, 1], const_z_percent=0, const_i_percent=0,
                 sn_mva=None,
                 name=None, scaling=1., index=[0, 1, 2])

    return net


def dtoc_relay_net(open_loop=True):
    net = create_empty_network()  # create an empty network

    # create buses
    create_buses(net, nr_buses=7, vn_kv=20, index=[0, 1, 2, 3, 4, 5, 6], name=None, type="n",
                 geodata=[(0, 0), (0, -1), (-2, -2), (-2, -4), (2, -2), (2, -3), (2, -4)])

    # external grids
    create_ext_grid(net, 0, vm_pu=1.0, va_degree=0, s_sc_max_mva=100, s_sc_min_mva=50, rx_max=0.1, rx_min=0.1)

    create_lines(net, from_buses=[0, 1, 2, 1, 4, 5, 3], to_buses=[1, 2, 3, 4, 5, 6, 6],
                 length_km=[2, 5, 4, 4, 0.5, 0.5, 1], std_type="NAYY 4x50 SE",
                 name=None, index=[0, 1, 2, 3, 4, 5, 6], df=1., parallel=1)

    net.line["endtemp_degree"] = 250

    create_switches(net, buses=[0, 1, 2, 3, 4, 5], elements=[0, 1, 2, 3, 4, 5], et='l', type="CB_DTOC")

    # Define switches
    if open_loop:
        create_switches(net, buses=[3, 6], elements=[6, 6], et='l', type="CB_DTOC", closed=False)
    else:
        create_switches(net, buses=[3, 6], elements=[6, 6], et='l', type="CB_DTOC", closed=True)

    #define load
    create_loads(net, buses=[3, 6], p_mw=[5, 2], q_mvar=[1, 1], const_z_percent=0, const_i_percent=0, sn_mva=None,
                 name=None, scaling=1., index=[0, 1])
    return net


def idmt_relay_net(open_loop=True):
    net = create_empty_network()  # create an empty network

    #create buses
    create_buses(net, nr_buses=7, vn_kv=20, index=[0, 1, 2, 3, 4, 5, 6], name=None, type="n",
                 geodata=[(0, 0), (0, -1), (-2, -2), (-2, -4), (2, -2), (2, -3), (2, -4)])

    # external grids
    create_ext_grid(net, 0, vm_pu=1.0, va_degree=0, s_sc_max_mva=100, s_sc_min_mva=50, rx_max=0.1, rx_min=0.1)

    create_lines(net, from_buses=[0, 1, 2, 1, 4, 5, 3], to_buses=[1, 2, 3, 4, 5, 6, 6],
                 length_km=[2, 5, 4, 4, 0.5, 0.5, 1], std_type="NAYY 4x50 SE",
                 name=None, index=[0, 1, 2, 3, 4, 5, 6], df=1., parallel=1)

    net.line["endtemp_degree"] = 250

    create_switches(net, buses=[0, 1, 2, 3, 4, 5], elements=[0, 1, 2, 3, 4, 5], et='l', type="CB_IDMT")

    # Define switches
    if open_loop:
        create_switches(net, buses=[3, 6], elements=[6, 6], et='l', type="CB_IDMT", closed=False)
    else:
        create_switches(net, buses=[3, 6], elements=[6, 6], et='l', type="CB_IDMT", closed=True)

    # define load
    create_loads(net, buses=[3, 6], p_mw=[5, 2], q_mvar=[1, 1], const_z_percent=0, const_i_percent=0, sn_mva=None,
                 name=None, scaling=1., index=[0, 1])
    return net


def idtoc_relay_net(open_loop=True):
    net = create_empty_network()  # create an empty network

    # create buses
    create_buses(net, nr_buses=7, vn_kv=20, index=[0, 1, 2, 3, 4, 5, 6], name=None, type="n",
                 geodata=[(0, 0), (0, -1), (-2, -2), (-2, -4), (2, -2), (2, -3), (2, -4)])

    # external grids
    create_ext_grid(net, 0, vm_pu=1.0, va_degree=0, s_sc_max_mva=100, s_sc_min_mva=50, rx_max=0.1, rx_min=0.1)

    create_lines(net, from_buses=[0, 1, 2, 1, 4, 5, 3], to_buses=[1, 2, 3, 4, 5, 6, 6],
                 length_km=[2, 5, 4, 4, 0.5, 0.5, 1], std_type="NAYY 4x50 SE",
                 name=None, index=[0, 1, 2, 3, 4, 5, 6], df=1., parallel=1)

    net.line["endtemp_degree"] = 250

    create_switches(net, buses=[0, 1, 2, 3, 4, 5], elements=[0, 1, 2, 3, 4, 5], et='l', type="CB_IDTOC")

    # Define switches
    if open_loop:
        create_switches(net, buses=[3, 6], elements=[6, 6], et='l', type="CB_IDTOC", closed=False)
    else:
        create_switches(net, buses=[3, 6], elements=[6, 6], et='l', type="CB_IDTOC", closed=True)

    # define load
    create_loads(net, buses=[3, 6], p_mw=[5, 2], q_mvar=[1, 1], const_z_percent=0, const_i_percent=0, sn_mva=None,
                 name=None, scaling=1., index=[0, 1])
    return net



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
from pandapower.auxiliary import _select_is_elements_numba, _check_connectivity_opf, _check_connectivity, \
    _set_isolated_buses_out_of_service, _replace_nans_with_default_limits, _write_lookup_to_net
from pandapower.build_branch import _switch_branches, _branches_with_oos_buses, \
    _build_branch_ppc, _build_tcsc_ppc, _build_branch_dc_ppc
from pandapower.build_bus import _build_bus_ppc, _calc_pq_elements_and_add_on_ppc, \
    _calc_shunts_and_add_on_ppc, _add_ext_grid_sc_impedance, _add_motor_impedances_ppc, \
    _build_svc_ppc, _add_load_sc_impedances_ppc, _build_ssc_ppc, _build_vsc_ppc, _build_bus_dc_ppc
from pandapower.build_gen import _build_gen_ppc, _check_voltage_setpoints_at_same_bus, \
    _check_voltage_angles_at_same_bus, _check_for_reference_bus
from pandapower.opf.make_objective import _make_objective
from pandapower.pypower.idx_area import PRICE_REF_BUS
from pandapower.pypower.idx_brch import F_BUS, T_BUS, BR_STATUS
from pandapower.pypower.idx_brch_dc import DC_F_BUS, DC_T_BUS, DC_BR_STATUS
from pandapower.pypower.idx_bus import NONE, BUS_I, BUS_TYPE, REF
from pandapower.pypower.idx_bus_dc import DC_BUS_I, DC_BUS_TYPE, DC_NONE, DC_B2B
from pandapower.pypower.idx_gen import GEN_BUS, GEN_STATUS
from pandapower.pypower.idx_ssc import SSC_STATUS, SSC_BUS, SSC_INTERNAL_BUS
from pandapower.pypower.idx_tcsc import TCSC_STATUS, TCSC_F_BUS, TCSC_T_BUS
from pandapower.pypower.idx_svc import SVC_STATUS, SVC_BUS
from pandapower.pypower.idx_vsc import VSC_BUS, VSC_INTERNAL_BUS, VSC_BUS_DC, VSC_STATUS, VSC_MODE_AC, VSC_MODE_AC_V, \
    VSC_MODE_AC_Q, VSC_MODE_AC_SL, VSC_INTERNAL_BUS_DC
from pandapower.pypower.run_userfcn import run_userfcn
from itertools import combinations


def _pd2ppc_recycle(net, sequence, recycle):
    # TODO: for FACTS and DC elements: SVC, TCSC, SSC, VSC
    key = "_ppc" if sequence is None else "_ppc%d" % sequence
    if not recycle or not net.get(key, None):
        return _pd2ppc(net, sequence=sequence)

    ppc = net[key]
    ppc["success"] = False
    ppc["iterations"] = 0.
    ppc["et"] = 0.

    if "bus_pq" in recycle and recycle["bus_pq"]:
        # update pq values in bus
        _calc_pq_elements_and_add_on_ppc(net, ppc, sequence=sequence)

    # if "trafo" in recycle and recycle["trafo"]:
    #     # update trafo in branch and Ybus
    #     lookup = net._pd2ppc_lookups["branch"]
    #     if "trafo" in lookup:
    #         _calc_trafo_parameter(net, ppc)
    #     if "trafo3w" in lookup:
    #         _calc_trafo3w_parameter(net, ppc)

    if "gen" in recycle and recycle["gen"]:
        # updates the ppc["gen"] part
        _build_gen_ppc(net, ppc)
        ppc["gen"] = np.nan_to_num(ppc["gen"])

    ppci = _ppc2ppci(ppc, net)
    ppci["internal"] = net[key]["internal"]
    net[key] = ppc

    return ppc, ppci


def _check_line_dc_at_b2b_buses(ppci):
    b2b_buses = ppci["bus_dc"][ppci["bus_dc"][:, DC_BUS_TYPE] == DC_B2B, DC_BUS_I].astype(np.int64)
    intersect_from = np.intersect1d(ppci["branch_dc"][:, DC_F_BUS].astype(np.int64), b2b_buses)
    intersect_to = np.intersect1d(ppci["branch_dc"][:, DC_T_BUS].astype(np.int64), b2b_buses)
    if len(intersect_from) != 0 or len(intersect_to) != 0:
        raise NotImplementedError("Found DC lines connected to Back-To-Back VSC converter configuration - "
                                  "not implemented. DC lines can only connect to the DC buses that are not "
                                  "part of a Back-To-Back configuration.")


def _check_vsc_different_ac_control_modes_at_same_bus(ppci):
    ac_vm_pu_buses = ppci["vsc"][ppci["vsc"][:, VSC_MODE_AC] == VSC_MODE_AC_V, VSC_BUS]
    ac_q_mvar_buses = ppci["vsc"][ppci["vsc"][:, VSC_MODE_AC] == VSC_MODE_AC_Q, VSC_BUS]
    ac_slack_buses = ppci["vsc"][ppci["vsc"][:, VSC_MODE_AC] == VSC_MODE_AC_SL, VSC_BUS]
    ac_bus_intersection = np.hstack(
        [np.intersect1d(a, b) for a, b in combinations([ac_vm_pu_buses, ac_q_mvar_buses, ac_slack_buses], r=2)])
    if len(ac_bus_intersection) != 0:
        raise NotImplementedError("Found multiple VSC converters that share the same AC bus and have "
                                  "different AC control modes - not implemented. VSC converters can only "
                                  "have the same AC control mode if they share the same AC bus.")


def _check_slack_at_vsc_bus(ppci):
    vsc_buses = ppci["vsc"][:, VSC_BUS]
    ac_slack_buses = ppci["gen"][ppci["internal"]["ref_gens"], GEN_BUS]
    ac_bus_intersection = np.intersect1d(vsc_buses, ac_slack_buses)
    if len(ac_bus_intersection) != 0:
        raise NotImplementedError("Found VSC elements connected to AC slack buses - "
                                  "this configuration is not implemented.")


def _pd2ppc(net, sequence=None, **kwargs):
    """
    Converter Flow:
        1. Create an empty pypower datatructure
        2. Calculate loads and write the bus matrix
        3. Build the gen (Infeeder)- Matrix
        4. Calculate the line parameter and the transformer parameter,
           and fill it in the branch matrix.
           Order: 1st: Line values, 2nd: Trafo values
        5. if opf: make opf objective (gencost)
        6. convert internal ppci format for pypower powerflow /
        opf without out of service elements and rearanged buses

    INPUT:
        **net** - The pandapower format network
        **sequence** - Used for three phase analysis
        ( 0 - Zero Sequence
          1 - Positive Sequence
          2 - Negative Sequence
        )

    OUTPUT:
        **ppc** - The simple matpower format network. Which consists of:
                  ppc = {
                        "baseMVA": 1., *float*
                        "version": 2,  *int*
                        "bus": np.array([], dtype=float),
                        "branch": np.array([], dtype=np.complex128),
                        "gen": np.array([], dtype=float),
                        "gencost" =  np.array([], dtype=float), only for OPF
                        "internal": {
                              "Ybus": np.array([], dtype=np.complex128)
                              , "Yf": np.array([], dtype=np.complex128)
                              , "Yt": np.array([], dtype=np.complex128)
                              , "branch_is": np.array([], dtype=bool)
                              , "gen_is": np.array([], dtype=bool)
                              }
        **ppci** - The "internal" pypower format network for PF calculations

    """
    # select elements in service (time consuming, so we do it once)
    net["_is_elements"] = _select_is_elements_numba(net, sequence=sequence)

    # Gets network configurations
    mode = net["_options"]["mode"]
    check_connectivity = net["_options"]["check_connectivity"]
    calculate_voltage_angles = net["_options"]["calculate_voltage_angles"]

    ppc = _init_ppc(net, mode=mode, sequence=sequence)

    # generate ppc['bus'] and the bus lookup
    _build_bus_ppc(net, ppc, sequence=sequence)
    _build_bus_dc_ppc(net, ppc)
    if sequence == 0:
        from pandapower.pd2ppc_zero import _add_ext_grid_sc_impedance_zero, _build_branch_ppc_zero
        # Adds external grid impedance for 3ph and sc calculations in ppc0
        _add_ext_grid_sc_impedance_zero(net, ppc)
        # Calculates ppc0 branch impedances from branch elements
        _build_branch_ppc_zero(net, ppc)
    else:
        # Calculates ppc1/ppc2 branch impedances from branch elements
        _build_branch_ppc(net, ppc)
    _build_branch_dc_ppc(net, ppc)

    _build_tcsc_ppc(net, ppc, mode)
    _build_svc_ppc(net, ppc, mode)
    _build_ssc_ppc(net, ppc, mode)
    _build_vsc_ppc(net, ppc, mode)

    # Adds P and Q for loads / sgens in ppc['bus'] (PQ nodes)
    if mode == "sc":
        _add_ext_grid_sc_impedance(net, ppc)
        # Generator impedance are seperately added in sc module
        _add_motor_impedances_ppc(net, ppc)
        if net._options.get("use_pre_fault_voltage", False):
            _add_load_sc_impedances_ppc(net, ppc)  # add SC impedances for loads

    else:
        _calc_pq_elements_and_add_on_ppc(net, ppc, sequence=sequence)
        # adds P and Q for shunts, wards and xwards (to PQ nodes)
        _calc_shunts_and_add_on_ppc(net, ppc)

    # adds auxilary buses for open switches at branches
    _switch_branches(net, ppc)

    # Adds auxilary buses for in service lines with out of service buses.
    # Also deactivates lines if they are connected to two out of service buses
    _branches_with_oos_buses(net, ppc)

    if check_connectivity:
        if sequence in [None, 1, 2]:
            # sets islands (multiple isolated nodes) out of service
            if mode == "opf":
                net["_isolated_buses"], _, _ = _check_connectivity_opf(ppc)
                net["_isolated_buses_dc"] = np.array([], dtype=np.int64)
            else:
                net["_isolated_buses"], _, _, net["_isolated_buses_dc"], _, _ = _check_connectivity(ppc)
            net["_is_elements_final"] = _select_is_elements_numba(net, net._isolated_buses,
                                                                  net._isolated_buses_dc, sequence)
        else:
            ppc["bus"][net._isolated_buses, BUS_TYPE] = NONE
        net["_is_elements"] = net["_is_elements_final"]
    else:
        # sets buses out of service, which aren't connected to branches / REF buses
        _set_isolated_buses_out_of_service(net, ppc)

    # we need to check this after checking connectivity (isolated vsc as DC slack cause change of DC_REF to DC_P)
    if "pf" in mode or "se" in mode:
        _check_for_reference_bus(ppc)

    _build_gen_ppc(net, ppc)

    _replace_nans_with_default_limits(net, ppc)

    # generates "internal" ppci format (for powerflow calc)
    # from "external" ppc format and updates the bus lookup
    # Note: Also reorders buses and gens in ppc
    ppci = _ppc2ppci(ppc, net)

    _check_line_dc_at_b2b_buses(ppci)
    _check_vsc_different_ac_control_modes_at_same_bus(ppci)
    # _check_slack_at_vsc_bus(ppci)

    if mode == "pf":
        # check if any generators connected to the same bus have different voltage setpoints
        _check_voltage_setpoints_at_same_bus(ppc)
        if calculate_voltage_angles:
            _check_voltage_angles_at_same_bus(net, ppci)

    if mode == "opf":
        # make opf objective
        ppci = _make_objective(ppci, net)

    return ppc, ppci


def _init_ppc(net, mode="pf", sequence=None):
    # init empty ppc
    ppc = {
        "baseMVA": net.sn_mva,
            "version": 2,
            "bus": np.array([], dtype=float),
            "bus_dc": np.array([], dtype=np.float64),
            "branch": np.array([], dtype=np.complex128),
            "branch_dc": np.array([], dtype=np.float64),
            "tcsc": np.array([], dtype=np.complex128),
            "svc": np.array([], dtype=np.complex128),
            "ssc": np.array([], dtype=np.complex128),
            "vsc": np.array([], dtype=np.float64),
            "gen": np.array([], dtype=float),
            "internal": {
                "Ybus": np.array([], dtype=np.complex128),
                "Yf": np.array([], dtype=np.complex128),
                "Yt": np.array([], dtype=np.complex128),
                "branch_is": np.array([], dtype=bool),
                "branch_dc_is": np.array([], dtype=bool),
                "gen_is": np.array([], dtype=bool),
                "DLF": np.array([], dtype=np.complex128),
                "buses_ord_bfs_nets": np.array([], dtype=float)
            }
        }
    if mode == "opf":
        # additional fields in ppc
        ppc["gencost"] = np.array([], dtype=float)
    net["_ppc"] = ppc

    if sequence is None:
        net["_ppc"] = ppc
    else:
        ppc["sequence"] = int(sequence)
        net["_ppc%s" % sequence] = ppc
    return ppc


def _ppc2ppci(ppc, net, ppci=None):
    """
    Creates the ppci which is used to run the power flow / OPF...
    The ppci is similar to the ppc except that:
    1. it contains no out of service elements
    2. buses are sorted

    Parameters
    ----------
    ppc - the ppc
    net - the pandapower net

    Returns
    -------
    ppci - the "internal" ppc

    """
    # get empty ppci
    if ppci is None:
        ppci = _init_ppc(net, mode=net["_options"]["mode"])
    # BUS Sorting and lookups
    # get bus_lookup
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    bus_dc_lookup = net["_pd2ppc_lookups"]["bus_dc"]
    # get OOS buses and place them at the end of the bus array
    # (there are no OOS busses in the ppci)
    oos_buses = ppc['bus'][:, BUS_TYPE] == NONE
    oos_buses_dc = ppc['bus_dc'][:, BUS_TYPE] == DC_NONE
    ppci['bus'] = ppc['bus'][~oos_buses]
    ppci['bus_dc'] = ppc['bus_dc'][~oos_buses_dc]
    # in ppc the OOS buses are included and at the end of the array
    ppc['bus'] = np.vstack([ppc['bus'][~oos_buses], ppc['bus'][oos_buses]])
    ppc['bus_dc'] = np.vstack([ppc['bus_dc'][~oos_buses_dc], ppc['bus_dc'][oos_buses_dc]])

    # generate bus_lookup_ppc_ppci (ppc -> ppci lookup)
    ppc_former_order = (ppc['bus'][:, BUS_I]).astype(np.int64)
    aranged_buses = np.arange(len(ppc["bus"]))

    # generate bus_dc_lookup_ppc_ppci (ppc -> ppci lookup)
    ppc_former_order_dc = (ppc['bus_dc'][:, DC_BUS_I]).astype(np.int64)
    aranged_buses_dc = np.arange(len(ppc["bus_dc"]))

    # lookup ppc former order -> consecutive order
    e2i = np.zeros(len(ppc["bus"]), dtype=np.int64)
    e2i[ppc_former_order] = aranged_buses

    # dc lookup ppc former order -> consecutive order
    e2i_dc = np.zeros(len(ppc["bus_dc"]), dtype=np.int64)
    e2i_dc[ppc_former_order_dc] = aranged_buses_dc

    # save consecutive indices in ppc and ppci
    ppc['bus'][:, BUS_I] = aranged_buses
    ppci['bus'][:, BUS_I] = ppc['bus'][:len(ppci['bus']), BUS_I]

    # save consecutive indices in ppc and ppci
    ppc['bus_dc'][:, DC_BUS_I] = aranged_buses_dc
    ppci['bus_dc'][:, DC_BUS_I] = ppc['bus_dc'][:len(ppci['bus_dc']), DC_BUS_I]

    # update lookups (pandapower -> ppci internal)
    _update_lookup_entries(net, bus_lookup, e2i, "bus")

    # update lookups (pandapower -> ppci internal)
    _update_lookup_entries(net, bus_dc_lookup, e2i_dc, "bus_dc")

    if 'areas' in ppc:
        if len(ppc["areas"]) == 0:  # if areas field is empty
            del ppc['areas']  # delete it (so it's ignored)

    # bus types
    bt = ppc["bus"][:, BUS_TYPE]
    bt_dc = ppc["bus_dc"][:, DC_BUS_TYPE]

    # update branch, gen and areas bus numbering
    ppc['gen'][:, GEN_BUS] = e2i[np.real(ppc["gen"][:, GEN_BUS]).astype(np.int64)].copy()
    ppc['svc'][:, SVC_BUS] = e2i[np.real(ppc["svc"][:, SVC_BUS]).astype(np.int64)].copy()
    ppc['ssc'][:, SSC_BUS] = e2i[np.real(ppc["ssc"][:, SSC_BUS]).astype(np.int64)].copy()
    ppc['ssc'][:, SSC_INTERNAL_BUS] = e2i[np.real(ppc["ssc"][:, SSC_INTERNAL_BUS]).astype(np.int64)].copy()
    ppc['vsc'][:, VSC_BUS] = e2i[np.real(ppc["vsc"][:, VSC_BUS]).astype(np.int64)].copy()
    ppc['vsc'][:, VSC_INTERNAL_BUS] = e2i[np.real(ppc["vsc"][:, VSC_INTERNAL_BUS]).astype(np.int64)].copy()
    ppc['vsc'][:, VSC_BUS_DC] = e2i_dc[np.real(ppc["vsc"][:, VSC_BUS_DC]).astype(np.int64)].copy()
    ppc['vsc'][:, VSC_INTERNAL_BUS_DC] = e2i_dc[np.real(ppc["vsc"][:, VSC_INTERNAL_BUS_DC]).astype(np.int64)].copy()
    ppc["branch"][:, F_BUS] = e2i[np.real(ppc["branch"][:, F_BUS]).astype(np.int64)].copy()
    ppc["branch"][:, T_BUS] = e2i[np.real(ppc["branch"][:, T_BUS]).astype(np.int64)].copy()
    ppc["branch_dc"][:, DC_F_BUS] = e2i_dc[np.real(ppc["branch_dc"][:, DC_F_BUS]).astype(np.int64)].copy()
    ppc["branch_dc"][:, DC_T_BUS] = e2i_dc[np.real(ppc["branch_dc"][:, DC_T_BUS]).astype(np.int64)].copy()
    ppc["tcsc"][:, TCSC_F_BUS] = e2i[np.real(ppc["tcsc"][:, TCSC_F_BUS]).astype(np.int64)].copy()
    ppc["tcsc"][:, TCSC_T_BUS] = e2i[np.real(ppc["tcsc"][:, TCSC_T_BUS]).astype(np.int64)].copy()

    # Note: The "update branch, gen and areas bus numbering" does the same as:
    # ppc['gen'][:, GEN_BUS] = get_indices(ppc['gen'][:, GEN_BUS], bus_lookup_ppc_ppci)
    # ppc["branch"][:, F_BUS] = get_indices(ppc["branch"][:, F_BUS], bus_lookup_ppc_ppci)
    # ppc["branch"][:, T_BUS] = get_indices( ppc["branch"][:, T_BUS], bus_lookup_ppc_ppci)
    # but faster...

    if 'areas' in ppc:
        ppc["areas"][:, PRICE_REF_BUS] = \
            e2i[np.real(ppc["areas"][:, PRICE_REF_BUS]).astype(np.int64)].copy()

    # initialize gen lookups
    for element, (f, t) in net._gen_order.items():
        _build_gen_lookups(net, element, f, t)

    # determine which buses, branches, gens are connected and
    # in-service
    n2i = ppc["bus"][:, BUS_I].astype(np.int64)
    bs = (bt != NONE)  # bus status

    n2i_dc = ppc["bus_dc"][:, DC_BUS_I].astype(np.int64)
    bs_dc = (bt_dc != DC_NONE)  # bus status

    gs = ((ppc["gen"][:, GEN_STATUS] > 0) &  # gen status
          bs[n2i[np.real(ppc["gen"][:, GEN_BUS]).astype(np.int64)]])
    ppci["internal"]["gen_is"] = gs

    svcs = ((ppc["svc"][:, SVC_STATUS] > 0) &  # gen status
            bs[n2i[np.real(ppc["svc"][:, SVC_BUS]).astype(np.int64)]])
    ppci["internal"]["svc_is"] = svcs

    sscs = ((ppc["ssc"][:, SSC_STATUS] > 0) &  # ssc status
            bs[n2i[np.real(ppc["ssc"][:, SSC_BUS]).astype(np.int64)]] &
            bs[n2i[np.real(ppc["ssc"][:, SSC_INTERNAL_BUS]).astype(np.int64)]])
    ppci["internal"]["ssc_is"] = sscs

    vscs = ((ppc["vsc"][:, VSC_STATUS] > 0) &  # vsc status
            bs[n2i[np.real(ppc["vsc"][:, VSC_BUS]).astype(np.int64)]] &
            bs[n2i[np.real(ppc["vsc"][:, VSC_INTERNAL_BUS]).astype(np.int64)]] &
            bs_dc[n2i_dc[np.real(ppc["vsc"][:, VSC_BUS_DC]).astype(np.int64)]] &
            bs_dc[n2i_dc[np.real(ppc["vsc"][:, VSC_INTERNAL_BUS_DC]).astype(np.int64)]])
    ppci["internal"]["vsc_is"] = vscs

    brs = (np.real(ppc["branch"][:, BR_STATUS]).astype(np.int64) &  # branch status
           bs[n2i[np.real(ppc["branch"][:, F_BUS]).astype(np.int64)]] &
           bs[n2i[np.real(ppc["branch"][:, T_BUS]).astype(np.int64)]]).astype(bool)
    ppci["internal"]["branch_is"] = brs

    brs_dc = (np.real(ppc["branch_dc"][:, DC_BR_STATUS]).astype(np.int64) &  # branch status
              bs_dc[n2i_dc[np.real(ppc["branch_dc"][:, DC_F_BUS]).astype(np.int64)]] &
              bs_dc[n2i_dc[np.real(ppc["branch_dc"][:, DC_T_BUS]).astype(np.int64)]]).astype(bool)
    ppci["internal"]["branch_dc_is"] = brs_dc

    trs = (np.real(ppc["tcsc"][:, TCSC_STATUS]).astype(np.int64) &  # branch status
           bs[n2i[np.real(ppc["tcsc"][:, TCSC_F_BUS]).astype(np.int64)]] &
           bs[n2i[np.real(ppc["tcsc"][:, TCSC_T_BUS]).astype(np.int64)]]).astype(bool)
    ppci["internal"]["tcsc_is"] = trs

    if 'areas' in ppc:
        ar = bs[n2i[ppc["areas"][:, PRICE_REF_BUS].astype(np.int64)]]
        # delete out of service areas
        ppci["areas"] = ppc["areas"][ar]

    # select in service elements from ppc and put them in ppci
    ppci["branch"] = ppc["branch"][brs]
    ppci["branch_dc"] = ppc["branch_dc"][brs_dc]
    ppci["tcsc"] = ppc["tcsc"][trs]

    ppci["gen"] = ppc["gen"][gs]
    ppci["svc"] = ppc["svc"][svcs]
    ppci["ssc"] = ppc["ssc"][sscs]
    ppci["vsc"] = ppc["vsc"][vscs]

    if 'dcline' in ppc:
        ppci['dcline'] = ppc['dcline']
    # execute userfcn callbacks for 'ext2int' stage
    if 'userfcn' in ppci:
        ppci = run_userfcn(ppci['userfcn'], 'ext2int', ppci)

    if net._pd2ppc_lookups["ext_grid"] is not None:
        ref_gens = np.setdiff1d(net._pd2ppc_lookups["ext_grid"], np.array([-1]))
    else:
        ref_gens = np.array([])
    if np.any(net.gen.slack.values[net._is_elements["gen"]]):
        slack_gens = np.array(net.gen.index)[net._is_elements["gen"]
                                             & net.gen["slack"].values]
        ref_gens = np.append(ref_gens, net._pd2ppc_lookups["gen"][slack_gens])
    ppci["internal"]["ref_gens"] = ref_gens.astype(np.int64)
    return ppci


def _update_lookup_entries(net, lookup, e2i, element):
    valid_bus_lookup_entries = lookup >= 0
    # update entries
    lookup[valid_bus_lookup_entries] = e2i[lookup[valid_bus_lookup_entries]]
    _write_lookup_to_net(net, element, lookup)


def _build_gen_lookups(net, element, f, t):
    in_service = net._is_elements[element]
    if "controllable" in element:
        pandapower_index = net[element.split("_")[0]].index.values[in_service]
    else:
        pandapower_index = net[element].index.values[in_service]
    ppc_index = np.arange(f, t)
    if len(pandapower_index) > 0:
        _init_lookup(net, element, pandapower_index, ppc_index)


def _init_lookup(net, lookup_name, pandapower_index, ppc_index):
    # init lookup
    lookup = -np.ones(max(pandapower_index) + 1, dtype=np.int64)

    # update lookup
    lookup[pandapower_index] = ppc_index

    _write_lookup_to_net(net, lookup_name, lookup)



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


"""Updates bus, gen, branch data structures to match power flow soln.
"""

from numpy import conj, zeros, complex128, abs, float64, sqrt, real, isin, arange
from numpy import finfo, c_, flatnonzero as find, setdiff1d, r_, int64

from pandapower.pypower.idx_brch import F_BUS, T_BUS, PF, PT, QF, QT
from pandapower.pypower.idx_bus import PD, QD
from pandapower.pypower.idx_gen import GEN_BUS, GEN_STATUS, PG, QG
from pandapower.pypower.idx_ssc import SSC_Q
from pandapower.pypower.idx_svc import SVC_Q
from pandapower.pypower.idx_vsc import VSC_P, VSC_Q
from pandapower.pypower.idx_tcsc import TCSC_QF, TCSC_QT, TCSC_PF, TCSC_PT
from pandapower.pypower.pfsoln import _update_v, _update_q, _update_p
from pandapower.auxiliary import version_check

try:
    from numba import jit
    version_check('numba')
except ImportError:
    from pandapower.pf.no_numba import jit

EPS = finfo(float).eps


def pfsoln(baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, Ybus, Yf, Yt, V, ref, ref_gens, Ibus=None,
           limited_gens=None):
    """Updates bus, gen, branch data structures to match power flow soln.

    @author: Ray Zimmerman (PSERC Cornell)
    @author: Richard Lincoln
    """
    # generator info
    on = find(gen[:, GEN_STATUS] > 0)  # which generators are on?
    gbus = gen[on, GEN_BUS].astype(int64)  # what buses are they at?

    # xward: add ref buses that are not at the generators
    xbus = setdiff1d(ref, gbus)

    # compute total injected bus powers
    Ibus = zeros(len(V)) if Ibus is None else Ibus
    Sbus = V * conj(Ybus * V - Ibus)

    _update_v(bus, V)
    # update gen results
    _update_q(baseMVA, bus, gen, gbus, Sbus[gbus], on)

    if limited_gens is not None and len(limited_gens) > 0:
        on = find((gen[:, GEN_STATUS] > 0) | isin(arange(len(gen)), limited_gens))
        gbus = gen[on, GEN_BUS].astype(int64)

    _update_p(baseMVA, bus, gen, ref, gbus, Sbus, ref_gens)

    # ----- update/compute branch power flows -----
    branch = _update_branch_flows(Yf, Yt, V, baseMVA, branch)

    return bus, gen, branch


def pf_solution_single_slack(baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, Ybus, Yf, Yt, V, ref, ref_gens,
                             Ibus=None, limited_gens=None):
    """
    faster version of pfsoln for a grid with a single slack bus

    NOTE: Do not use in combination with shunts (check if ppc["bus"][:, GS/BS] are != 0.)
    NOTE: Do not use in combination with voltage dependend loads

    """

    # ----- update bus voltages -----
    _update_v(bus, V)

    # ----- update/compute branch power flows -----
    branch = _update_branch_flows(Yf, Yt, V, baseMVA, branch)

    p_bus = bus[:, PD].sum()
    q_bus = bus[:, QD].sum()
    p_loss = branch[:, [PF, PT]].sum()
    q_loss = branch[:, [QF, QT]].sum()

    # consider FACTS devices:
    q_facts = svc[:, SVC_Q].sum() + tcsc[:, [TCSC_QF, TCSC_QT]].sum() + ssc[:, SSC_Q].sum() + vsc[:, VSC_Q].sum()
    p_facts = tcsc[:, [TCSC_PF, TCSC_PT]].sum() + vsc[:, VSC_P].sum()  # now should all be 0 due to no losses

    # slack p = sum of branch losses and p demand at all buses
    gen[:, PG] = p_loss.real + p_bus + p_facts  # branch p losses + p demand
    gen[:, QG] = q_loss.real + q_bus + q_facts  # branch q losses + q demand

    return bus, gen, branch


def _update_branch_flows(Yf, Yt, V, baseMVA, branch):
    f_bus = real(branch[:, F_BUS]).astype(int64)
    t_bus = real(branch[:, T_BUS]).astype(int64)
    # complex power at "from" bus
    Sf = calc_branch_flows(Yf.data, Yf.indptr, Yf.indices, V, baseMVA, Yf.shape[0], f_bus)
    # complex power injected at "to" bus
    St = calc_branch_flows(Yt.data, Yt.indptr, Yt.indices, V, baseMVA, Yt.shape[0], t_bus)
    branch[:, [PF, QF, PT, QT]] = c_[Sf.real, Sf.imag, St.real, St.imag]
    return branch


@jit(nopython=True, cache=False)
def calc_branch_flows(Yy_x, Yy_p, Yy_j, v, baseMVA, dim_x, bus_ind):  # pragma: no cover

    Sx = zeros(dim_x, dtype=complex128)

    # iterate through sparse matrix and get Sx = conj(Y_kj* V[j])
    for r in range(len(Yy_p) - 1):
        for k in range(Yy_p[r], Yy_p[r + 1]):
            Sx[r] += conj(Yy_x[k] * v[Yy_j[k]]) * baseMVA

    # finally get Sx = V[k] * conj(Y_kj* V[j])
    Sx *= v[bus_ind]

    return Sx


@jit(nopython=True, cache=False)
def calc_branch_flows_batch(Yy_x, Yy_p, Yy_j, V, baseMVA, dim_x, bus_ind, base_kv):  # pragma: no cover
    """
    Function to get branch flows with a batch computation for the timeseries module

    Parameters
    ----------
    Yy_x, Yy_p, Yy_j - Yt or Yf CSR represenation
    V - complex voltage matrix results from time series
    baseMVA - base MVA from ppc
    dim_x - shape of Y
    bus_ind - f_bus or t_bus
    base_kv - pcci["bus"] BASE_KV values

    Returns
    ----------
    i_abs, s_abs - absolute branch currents and power flows. This is "i_ft" / "s_ft" in results_branch.py
    S - complex Sf / St values frpm ppci

    """

    S = zeros((V.shape[0], dim_x), dtype=complex128)
    s_abs = zeros((V.shape[0], dim_x), dtype=float64)
    i_abs = zeros((V.shape[0], dim_x), dtype=float64)
    sqrt_3 = sqrt(3)

    # iterate over entries in V (v= complex V result of each time step)
    for t in range(V.shape[0]):
        v = V[t]
        vm = abs(v)
        Sx = zeros(dim_x, dtype=complex128)

        # iterate through sparse matrix and get Sx = conj(Y_kj* V[j])
        for r in range(len(Yy_p) - 1):
            for k in range(Yy_p[r], Yy_p[r + 1]):
                Sx[r] += conj(Yy_x[k] * v[Yy_j[k]]) * baseMVA

        # finally get Sx = V[k] * conj(Y_kj* V[j])
        Sx *= v[bus_ind]
        S[t, :] = Sx
        s_abs[t, :] = abs(Sx)
        i_abs[t, :] = s_abs[t, :] / (vm[bus_ind] * base_kv[bus_ind]) / sqrt_3

    return S, s_abs, i_abs



import numbers
import numpy as np

from pandapower.control.basic_controller import Controller
from pandapower.auxiliary import _detect_read_write_flag, read_from_net, write_to_net


class BinarySearchControl(Controller):
    """
        The Binary search control is a controller which is used to reach a given set point . It can be used for
        reactive power control or voltage control. in case of voltage control, the input parameter voltage_ctrl must be
        set to true. Input and output elements and indexes can be lists. Input elements can be transformers, switches,
        lines or busses (only in case of voltage control). in case of voltage control, a bus_index must be present,
        where the voltage will be controlled. Output elements are sgens, where active and reactive power can be set. The
        output value distribution describes the distribution of reactive power provision between multiple
        output_elements and must sum up to 1.

        INPUT:
            **self**

            **net** - A pandapower grid

            **ctrl_in_service** - Whether the controller is in service or not.

            **output_element** - Output element of the controller. Takes a string value "gen" or "sgen", with
            reactive power control, currently only "sgen" is possible.

            **output_variable** - Output variable of that element, normally "q_mvar".

            **output_element_index** - Index of output element in e.g. "net.sgen".

            **output_element_in_service** - Whether output elements are in service or not.

            **output_values_distribution** - Distribution of reactive power provision.

            **input_element** - Measurement location, can be a transformers, switches, lines or busses (only with
            voltage_ctrl), indicated by string value "res_trafo", "res_switch", "res_line" or "res_bus". In case of
            "res_switch", an additional small impedance is introduced in the switch.

            **input_variable** - Variable which is used to take the measurement from. Indicated by string value.

            **input_element_index** - Element of input element in net.

            **set_point** - Set point of the controller, can be a reactive power provision or a voltage set point. In
            case of voltage set point, voltage control must be set to true, bus_idx must be set to measurement bus and
            input_element must be "res_bus". Can be overwritten by a droop controller chained with the binary search
            control.

            **voltage_ctrl** - Whether the controller is used for voltage control or not.

            **bus_idx=None** - Bus index which is used for voltage control.

            **tol=0.001** - Tolerance criteria of controller convergence.
       """
    def __init__(self, net, ctrl_in_service, output_element, output_variable, output_element_index,
                 output_element_in_service, output_values_distribution, input_element, input_variable,
                 input_element_index, set_point, voltage_ctrl, bus_idx=None, tol=0.001, in_service=True, order=0, level=0,
                 drop_same_existing_ctrl=False, matching_params=None, **kwargs):
        super().__init__(net, in_service=in_service, order=order, level=level,
                         drop_same_existing_ctrl=drop_same_existing_ctrl,
                         matching_params=matching_params, **kwargs)
        self.in_service = ctrl_in_service
        self.input_element = input_element
        self.input_element_index = []
        if isinstance(input_element_index, list):
            for element in input_element_index:
                self.input_element_index.append(element)
        else:
            self.input_element_index.append(input_element_index)
        self.output_element = output_element
        self.output_element_index = output_element_index
        self.output_element_in_service = output_element_in_service
        # normalize the values distribution:
        self.output_values_distribution = np.array(output_values_distribution, dtype=np.float64) / np.sum(
            output_values_distribution)
        self.set_point = set_point
        self.voltage_ctrl = voltage_ctrl
        self.bus_idx = bus_idx
        self.tol = tol
        self.applied = False
        self.output_values = None
        self.output_values_old = None
        self.diff = None
        self.diff_old = None
        self.converged = False
        self.overwrite_covergence = False
        self.write_flag, self.output_variable = _detect_read_write_flag(net, output_element, output_element_index,
                                                                        output_variable)
        self.read_flag = []
        self.input_variable = []
        self.input_element_in_service = []
        counter = 0
        for input_index in self.input_element_index:
            if self.input_element == "res_line":
                self.input_element_in_service.append(net.line.in_service[input_index])
            elif self.input_element == "res_trafo":
                self.input_element_in_service.append(net.trafo.in_service[input_index])
            elif self.input_element == "res_switch":
                self.input_element_in_service.append(
                    net[self.input_element].pf_in_service[input_index])
            elif self.input_element == "res_bus":
                self.input_element_in_service.append(net.bus.in_service[input_index])

            if isinstance(input_variable, list):
                read_flag_temp, input_variable_temp = _detect_read_write_flag(net, self.input_element,
                                                                              input_index,
                                                                              input_variable[counter])
            else:
                read_flag_temp, input_variable_temp = _detect_read_write_flag(net, self.input_element,
                                                                              input_index,
                                                                              input_variable)
            self.read_flag.append(read_flag_temp)
            self.input_variable.append(input_variable_temp)
            counter += 1

    def initialize_control(self, net):
        self.output_values = read_from_net(net, self.output_element, self.output_element_index, self.output_variable,
                                           self.write_flag)

    def is_converged(self, net):
        """
        Actual implementation of the convergence criteria: If controller is applied, it can stop
        """
        # if controller not in_service, return True
        self.in_service = net.controller.in_service[self.index]
        if not self.in_service:
            return True
        self.input_element_in_service.clear()
        self.output_element_in_service.clear()
        for input_index in self.input_element_index:
            if self.input_element == "res_line":
                self.input_element_in_service.append(net.line.in_service[input_index])
            elif self.input_element == "res_trafo":
                self.input_element_in_service.append(net.trafo.in_service[input_index])
            elif self.input_element == "res_switch":
                self.input_element_in_service.append(net.switch.closed[input_index])
            elif self.input_element == "res_bus":
                self.input_element_in_service.append(net.bus.in_service[input_index])
        for output_index in self.output_element_index:
            if self.output_element == "gen":
                self.output_element_in_service.append(net.gen.in_service[output_index])
            elif self.output_element == "sgen":
                self.output_element_in_service.append(net.sgen.in_service[output_index])
            elif self.output_element == "shunt":
                self.output_element_in_service.append(net.shunt.in_service[output_index])
        # check if at least one input and one output element is in_service
        if not (any(self.input_element_in_service) and any(self.output_element_in_service)):
            self.converged = True
            return self.converged

        # read input values
        input_values = []
        counter = 0
        for input_index in self.input_element_index:
            input_values.append(read_from_net(net, self.input_element, input_index,
                                              self.input_variable[counter], self.read_flag[counter]))
            counter += 1
        # read previous set values
        # compare old and new set values
        if not self.voltage_ctrl or self.bus_idx is None:
            self.diff_old = self.diff
            self.diff = self.set_point - sum(input_values)
            self.converged = np.all(np.abs(self.diff) < self.tol)
        else:
            self.diff_old = self.diff
            self.diff = self.set_point - net.res_bus.vm_pu.at[self.bus_idx]
            self.converged = np.all(np.abs(self.diff) < self.tol)

        if self.overwrite_covergence:
            self.overwrite_covergence = False
            return False
        else:
            return self.converged

    def control_step(self, net):
        self._binarysearchcontrol_step(net)

    def _binarysearchcontrol_step(self, net):
        if not self.in_service:
            return
        if self.output_values_old is None:
            self.output_values_old, self.output_values = self.output_values, self.output_values + 1e-3
        else:
            step_diff = self.diff - self.diff_old
            x = self.output_values - self.diff * (self.output_values - self.output_values_old) / np.where(
                step_diff == 0, 1e-6, step_diff)
            x = x * self.output_values_distribution if isinstance(x, numbers.Number) else sum(
                x) * self.output_values_distribution
            self.output_values_old, self.output_values = self.output_values, x

        # write new set values
        write_to_net(net, self.output_element, self.output_element_index, self.output_variable, self.output_values,
                     self.write_flag)

    def __str__(self):
        return super().__str__() + " [%s.%s.%s.%s]" % (
            self.input_element, self.input_variable, self.output_element, self.output_variable)


class DroopControl(Controller):
    """
            The droop controller is used in case of a droop based control. It can operate either as a Q(U) controller or
            as a U(Q) controller and is used in addition to a binary search controller (bsc). The linked binary search
            controller is specified using the controller index, which refers to the linked bsc. The droop controller
            behaves in a similar way to the station controllers presented in the Power Factory Tech Ref, although not
            all possible settings from Power Factory are yet available.

            INPUT:
                **self**

                **net** - A pandapower grid.

                **q_droop_var** - Droop Value in Mvar/p.u.

                **bus_idx** - Bus index in case of voltage control.

                **vm_set_pu** - Voltage set point in case of voltage control.

                **controller_idx** - Index of linked Binary< search control (if present).

                **voltage_ctrl** - Whether the controller is used for voltage control or not.

                **bus_idx=None** - Bus index which is used for voltage control.

                **tol=1e-6** - Tolerance criteria of controller convergence.

                **vm_set_lb=None** - Lower band border of dead band

                **vm_set_ub=None** - Upper band border of dead band
           """
    def __init__(self, net, q_droop_mvar, bus_idx, vm_set_pu, controller_idx, voltage_ctrl, tol=1e-6, in_service=True,
                 order=-1, level=0, drop_same_existing_ctrl=False, matching_params=None, vm_set_lb=None, vm_set_ub=None,
                 **kwargs):
        super().__init__(net, in_service=in_service, order=order, level=level,
                         drop_same_existing_ctrl=drop_same_existing_ctrl,
                         matching_params=matching_params, **kwargs)
        # TODO: implement maximum and minimum of droop control
        self.q_droop_mvar = q_droop_mvar
        self.bus_idx = bus_idx
        self.vm_pu = None
        self.vm_pu_old = self.vm_pu
        self.vm_set_pu = vm_set_pu
        self.vm_set_pu_new = None
        self.lb_voltage = vm_set_lb
        self.ub_voltage = vm_set_ub
        self.controller_idx = controller_idx
        self.voltage_ctrl = voltage_ctrl
        self.tol = tol
        self.applied = False
        self.read_flag, self.input_variable = _detect_read_write_flag(net, "res_bus", bus_idx, "vm_pu")
        self.q_set_mvar_bsc = None
        self.q_set_mvar = None
        self.q_set_old_mvar = None
        self.diff = None
        self.converged = False

    def is_converged(self, net):
        if self.voltage_ctrl:
            self.diff = (net.controller.at[self.controller_idx, "object"].set_point -
                         read_from_net(net, "res_bus", self.bus_idx, "vm_pu", self.read_flag))
        else:
            counter = 0
            input_values = []
            for input_index in net.controller.at[self.controller_idx, "object"].input_element_index:
                input_values.append(
                    read_from_net(net, net.controller.at[self.controller_idx, "object"].input_element, input_index,
                                  net.controller.at[self.controller_idx, "object"].input_variable[counter],
                                  net.controller.at[self.controller_idx, "object"].read_flag[counter]))
                counter += 1
            self.diff = ((net.controller.at[self.controller_idx, "object"].set_point - sum(input_values)))
        # bigger differences with switches as input elements, increase tolerance
        #if net.controller.at[self.controller_idx, "object"].input_element == "res_switch":
        #    self.tol = 0.2
        if self.bus_idx is None:
            self.converged = np.all(np.abs(self.diff) < self.tol)
        else:
            if np.all(np.abs(self.diff) < self.tol):
                self.converged = net.controller.at[self.controller_idx, "object"].converged
            elif net.controller.at[self.controller_idx, "object"].diff_old is not None:
                net.controller.at[self.controller_idx, "object"].overwrite_covergence = True

        return self.converged

    def control_step(self, net):
        self._droopcontrol_step(net)

    def _droopcontrol_step(self, net):
        self.vm_pu_old = self.vm_pu
        self.vm_pu = read_from_net(net, "res_bus", self.bus_idx, "vm_pu", self.read_flag)
        if not self.voltage_ctrl:
            if self.q_set_mvar_bsc is None:
                self.q_set_mvar_bsc = net.controller.at[self.controller_idx, "object"].set_point
            if self.lb_voltage is not None and self.ub_voltage is not None:
                if self.vm_pu > self.ub_voltage:
                    self.q_set_old_mvar, self.q_set_mvar = (
                        self.q_set_mvar, self.q_set_mvar_bsc + (self.ub_voltage - self.vm_pu) * self.q_droop_mvar)
                elif self.vm_pu < self.lb_voltage:
                    self.q_set_old_mvar, self.q_set_mvar = (
                        self.q_set_mvar, self.q_set_mvar_bsc + (self.lb_voltage - self.vm_pu) * self.q_droop_mvar)
                else:
                    self.q_set_old_mvar, self.q_set_mvar = (self.q_set_mvar, self.q_set_mvar_bsc)
            else:
                self.q_set_old_mvar, self.q_set_mvar = (
                    self.q_set_mvar, self.q_set_mvar - (self.vm_set_pu - self.vm_pu) * self.q_droop_mvar)

            if self.q_set_old_mvar is not None:
                self.diff = self.q_set_mvar - self.q_set_old_mvar
            if self.q_set_mvar is not None:
                net.controller.at[self.controller_idx, "object"].set_point = self.q_set_mvar

        else:
            input_element = net.controller.at[self.controller_idx, "object"].input_element
            input_element_index = net.controller.at[self.controller_idx, "object"].input_element_index
            input_variable = net.controller.at[self.controller_idx, "object"].input_variable
            read_flag = net.controller.at[self.controller_idx, "object"].read_flag
            input_values = []
            counter = 0
            for input_index in input_element_index:
                input_values.append(read_from_net(net, input_element, input_index,
                                                  input_variable[counter], read_flag[counter]))
            self.vm_set_pu_new = self.vm_set_pu + sum(input_values) / self.q_droop_mvar
            net.controller.at[self.controller_idx, "object"].set_point = self.vm_set_pu_new



from pandapower.converter.matpower import *
from pandapower.converter.pypower import *
from pandapower.converter.pandamodels import *
from pandapower.converter.ucte import *
from pandapower.converter.cim import *
from pandapower.converter.powerfactory import *
from pandapower.converter.jao import *



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import logging
import time
from typing import Dict
import pandas as pd
import numpy as np
from pandapower.auxiliary import pandapowerNet
from .. import cim_tools


class CreateMeasurements:

    def __init__(self, net: pandapowerNet, cim: Dict):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.net = net
        self.cim = cim

    def _set_measurement_element_datatype(self):
        self.net.measurement.element = self.net.measurement.element.astype(np.sctypeDict.get("UInt32"))

    def _copy_to_measurement(self, input_df: pd.DataFrame):
        pp_type = 'measurement'
        self.logger.debug("Copy %s datasets to pandapower network with type %s" % (input_df.index.size, pp_type))
        if pp_type not in self.net.keys():
            self.logger.warning("Missing pandapower type %s in the pandapower network!" % pp_type)
            return
        if input_df.empty:
            return
        self.net[pp_type] = pd.concat([self.net[pp_type],
                                      input_df[list(set(self.net[pp_type].columns).intersection(input_df.columns))]],
                                      ignore_index=True, sort=False)

    def create_measurements_from_analog(self):
        self.logger.info("------------------------- Creating measurements from Analog -------------------------")
        time_start = time.time()
        sc = cim_tools.get_pp_net_special_columns_dict()
        # join the Analogs with the AnalogValues and MeasurementValueSources
        analogs_prf = 'op' if 'op' in self.cim.keys() else 'eq'
        analogs = pd.merge(
            self.cim[analogs_prf]['Analog'][['rdfId', 'name', 'description', 'measurementType', 'unitSymbol',
                                             'unitMultiplier', 'Terminal', 'PowerSystemResource', 'positiveFlowIn']],
            self.cim[analogs_prf]['AnalogValue'][['rdfId', 'sensorAccuracy', 'MeasurementValueSource', 'Analog',
                                                  'value']],
            how='inner', left_on='rdfId', right_on='Analog', suffixes=("_Analog", "_AnalogValue"))
        analogs = analogs.drop(columns=['Analog'])
        analogs = analogs.rename(columns={'name': 'name_analog'})
        analogs = pd.merge(analogs, self.cim[analogs_prf]['MeasurementValueSource'], how='left',
                           left_on='MeasurementValueSource',
                           right_on='rdfId')
        analogs = analogs.drop(columns=['rdfId', 'MeasurementValueSource'])
        analogs = analogs.rename(columns={'name': sc['src']})
        # collect all the assets (line, trafo, trafo3w) and its connections
        assets = pd.DataFrame(None, columns=['element_type', 'side'])
        append_dict = dict({'line': {'from_bus': 'from', 'to_bus': 'to'},
                            'trafo': {'hv_bus': 'hv', 'lv_bus': 'lv'},
                            'trafo3w': {'hv_bus': 'hv', 'mv_bus': 'mv', 'lv_bus': 'lv'},
                            'load': {'bus': None},
                            'sgen': {'bus': None},
                            'gen': {'bus': None},
                            'shunt': {'bus': None},
                            'ext_grid': {'bus': None},
                            'ward': {'bus': None},
                            'xward': {'bus': None}
                            })
        for element_type, sides in append_dict.items():
            for side_name, side in sides.items():
                temp = self.net[element_type][[sc['o_id'], side_name, sc[side_name]]]. \
                    reset_index().rename(columns={side_name: 'bus', sc[side_name]: 'terminal_asset'})
                temp['element_type'] = element_type
                temp['side'] = side
                assets = pd.concat([assets, temp], sort=False)
        assets = assets.rename(columns={'index': 'element'})
        # now join the analogs with the assets
        psr = pd.merge(analogs, assets, how='inner', left_on='PowerSystemResource', right_on=sc['o_id'])
        # keep only entries which are associated to the terminal from the asset
        psr = psr.loc[psr.Terminal == psr.terminal_asset]
        # remove the PhaseVoltage measurements
        psr = psr.loc[psr.measurementType != 'PhaseVoltage']
        psr['measurement_type'] = psr.unitSymbol.map({'W': 'p', 'VAr': 'q', 'A': 'i', 'V': 'v'})
        # change the sign if needed
        psr['value'].loc[~psr['positiveFlowIn']] = psr.loc[~psr['positiveFlowIn']]['value'] * (-1)
        # move the voltage measurements to the buses
        psr = pd.merge(psr, self.net.bus[['vn_kv']], how='inner', left_on='bus', right_index=True)
        psr_v = psr.loc[psr['measurement_type'] == 'v']     
        psr_v['value'] = psr_v['value'] / psr_v['vn_kv']
        psr_v['std_dev'] = psr_v['sensorAccuracy'] / psr_v['vn_kv']
        psr_v['element_type'] = 'bus'
        psr_v['element'] = psr_v['bus']
        psr_v['side'] = None
        psr.loc[psr['measurement_type'] == 'v', psr_v.columns] = psr_v

        # convert amps to ka and assign std_dev values for 'i' measurements
        psr_i = psr.loc[psr['measurement_type'] == 'i'] 
        psr_i['value'] = psr_i['value'] / 1e3
        psr_i['std_dev'] = psr_i['sensorAccuracy'] / 1e3
        psr.loc[psr['measurement_type'] == 'i', psr_i.columns] = psr_i

        # assign std_dev values for 'p' measurements
        psr_p = psr.loc[psr['measurement_type'] == 'p'] 
        psr_p['std_dev'] = psr_p['sensorAccuracy'] 
        psr.loc[psr['measurement_type'] == 'p', psr_p.columns] = psr_p

        # assign std_dev values for 'q' measurements
        psr_q = psr.loc[psr['measurement_type'] == 'q'] 
        psr_q['std_dev'] = psr_q['sensorAccuracy'] 
        psr.loc[psr['measurement_type'] == 'q', psr_q.columns] = psr_q

        psr = psr.drop(columns=[sc['o_id']])
        psr = psr.rename(columns={'rdfId_AnalogValue': sc['o_id'], 'name_analog': sc['name'],
                                  'rdfId_Analog': sc['a_id'], 'Terminal': 'terminal_id'})
        psr[sc['o_cl']] = 'AnalogValue'
        psr = psr.reset_index(drop=True)

        self._copy_to_measurement(psr)

        # set the element from float to default uint32
        self._set_measurement_element_datatype()

        self.logger.info("Needed time for creating the measurements: %ss" % (time.time() - time_start))

    def create_measurements_from_sv(self):
        self.logger.info("--------------------------- Creating measurements from SV ---------------------------")
        time_start = time.time()
        sc = cim_tools.get_pp_net_special_columns_dict()
        # get the measurements from the sv profile and set the Terminal as index
        sv_powerflow = self.cim['sv']['SvPowerFlow'][['rdfId', 'Terminal', 'p', 'q']]
        sv_powerflow = sv_powerflow.set_index('Terminal')
        sv_powerflow[sc['o_cl']] = 'SvPowerFlow'
        sv_powerflow[sc['src']] = 'SV'
        sv_powerflow[sc['desc']] = None
        sv_powerflow[sc['a_id']] = None
        sv_powerflow = sv_powerflow.rename(columns={'rdfId': sc['o_id']})

        # ---------------------------------------measure: bus v---------------------------------------------------
        busses_temp = self.net.bus[['name', 'vn_kv', sc['ct']]].copy()
        busses_temp = busses_temp.reset_index(level=0)
        busses_temp = busses_temp.rename(columns={'index': 'element', sc['ct']: 'TopologicalNode'})
        sv_sv_voltages = pd.merge(self.cim['sv']['SvVoltage'][['rdfId', 'TopologicalNode', 'v']], busses_temp,
                                  how='left', on='TopologicalNode')
        # drop all the rows mit vn_kv == np.nan (no measurements available for that bus)
        sv_sv_voltages = sv_sv_voltages.dropna(subset=['vn_kv'])
        sv_sv_voltages = sv_sv_voltages.reset_index()
        if 'index' in sv_sv_voltages.columns:
            sv_sv_voltages = sv_sv_voltages.drop(['index'], axis=1)
        # value -> voltage ()
        sv_sv_voltages['value'] = sv_sv_voltages.v / sv_sv_voltages.vn_kv
        sv_sv_voltages['value'] = sv_sv_voltages['value'].replace(0, np.nan)
        # drop all the rows mit value == np.nan
        sv_sv_voltages = sv_sv_voltages.dropna(subset=['value'])
        sv_sv_voltages = sv_sv_voltages.reset_index()
        sv_sv_voltages['value_stddev'] = sv_sv_voltages.value * 0.001
        sv_sv_voltages['vn_kv_stddev'] = 0.1 / sv_sv_voltages.vn_kv
        sv_sv_voltages['std_dev'] = sv_sv_voltages[["value_stddev", "vn_kv_stddev"]].max(axis=1)
        sv_sv_voltages['measurement_type'] = 'v'
        sv_sv_voltages['element_type'] = 'bus'
        sv_sv_voltages['side'] = None
        sv_sv_voltages[sc['src']] = 'SV'
        sv_sv_voltages[sc['o_cl']] = 'SvVoltage'
        sv_sv_voltages[sc['desc']] = None
        sv_sv_voltages[sc['a_id']] = None
        sv_sv_voltages = sv_sv_voltages.rename(columns={'rdfId': sc['o_id']})

        self._copy_to_measurement(sv_sv_voltages)

        # ---------------------------------------measure: line---------------------------------------------------
        sigma_line = 0.03
        line_temp = self.net.line[['name', 'from_bus', 'to_bus', sc['t_from'], sc['t_to']]].copy()
        line_temp['p_from'] = \
            pd.merge(line_temp[sc['t_from']], sv_powerflow['p'], left_on=sc['t_from'], right_index=True)['p']
        line_temp['p_to'] = \
            pd.merge(line_temp[sc['t_to']], sv_powerflow['p'], left_on=sc['t_to'], right_index=True)['p']
        line_temp['q_from'] = \
            pd.merge(line_temp[sc['t_from']], sv_powerflow['q'], left_on=sc['t_from'], right_index=True)['q']
        line_temp['q_to'] = \
            pd.merge(line_temp[sc['t_to']], sv_powerflow['q'], left_on=sc['t_to'], right_index=True)['q']

        line_temp = line_temp.dropna(subset=['p_from', 'p_to', 'q_from', 'q_to'], thresh=4)

        line_temp['stddev_line_from_p'] = abs(line_temp.p_from) * sigma_line + 1.
        line_temp['stddev_line_to_p'] = abs(line_temp.p_to) * sigma_line + 1.
        line_temp['stddev_line_from_q'] = abs(line_temp.q_from) * sigma_line + 1.
        line_temp['stddev_line_to_q'] = abs(line_temp.q_to) * sigma_line + 1.
        line_temp['element_type'] = 'line'
        line_temp['element'] = line_temp.index[:]
        # copy the data into the measurement dataframe
        # ---------------------------------------measure: line p from---------------------------------------------------
        line_temp['measurement_type'] = 'p'
        line_temp['value'] = line_temp.p_from
        line_temp['std_dev'] = line_temp.stddev_line_from_p
        line_temp['side'] = line_temp.from_bus
        line_temp = pd.merge(line_temp, sv_powerflow, left_on=sc['t_from'], right_index=True)
        self._copy_to_measurement(line_temp)
        line_temp = line_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: line p to---------------------------------------------------
        line_temp['value'] = line_temp.p_to
        line_temp['std_dev'] = line_temp.stddev_line_to_p
        line_temp['side'] = line_temp.to_bus
        line_temp = pd.merge(line_temp, sv_powerflow, left_on=sc['t_to'], right_index=True)
        self._copy_to_measurement(line_temp)
        line_temp = line_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: line q from---------------------------------------------------
        line_temp['measurement_type'] = 'q'
        line_temp['value'] = line_temp.q_from
        line_temp['std_dev'] = line_temp.stddev_line_from_q
        line_temp['side'] = line_temp.from_bus
        line_temp = pd.merge(line_temp, sv_powerflow, left_on=sc['t_from'], right_index=True)
        self._copy_to_measurement(line_temp)
        line_temp = line_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: line q to---------------------------------------------------
        line_temp['value'] = line_temp.q_to
        line_temp['std_dev'] = line_temp.stddev_line_to_q
        line_temp['side'] = line_temp.to_bus
        line_temp = pd.merge(line_temp, sv_powerflow, left_on=sc['t_to'], right_index=True)
        self._copy_to_measurement(line_temp)

        # ---------------------------------------measure: trafo---------------------------------------------------
        sigma_trafo = 0.03
        trafo_temp = self.net.trafo[['name', 'hv_bus', 'lv_bus', sc['t_hv'], sc['t_lv']]].copy()
        # if trafo_temp.index.size > 0:
        trafo_temp['p_hv'] = \
            pd.merge(trafo_temp[sc['t_hv']], sv_powerflow['p'], left_on=sc['t_hv'], right_index=True)['p']
        trafo_temp['p_lv'] = \
            pd.merge(trafo_temp[sc['t_lv']], sv_powerflow['p'], left_on=sc['t_lv'], right_index=True)['p']
        trafo_temp['q_hv'] = \
            pd.merge(trafo_temp[sc['t_hv']], sv_powerflow['q'], left_on=sc['t_hv'], right_index=True)['q']
        trafo_temp['q_lv'] = \
            pd.merge(trafo_temp[sc['t_lv']], sv_powerflow['q'], left_on=sc['t_lv'], right_index=True)['q']

        trafo_temp = trafo_temp.dropna(subset=['p_hv', 'p_lv', 'q_hv', 'q_lv'], thresh=4)

        trafo_temp['stddev_trafo_hv_p'] = abs(trafo_temp.p_hv) * sigma_trafo + 1.
        trafo_temp['stddev_trafo_lv_p'] = abs(trafo_temp.p_lv) * sigma_trafo + 1.
        trafo_temp['stddev_trafo_hv_q'] = abs(trafo_temp.q_hv) * sigma_trafo + 1.
        trafo_temp['stddev_trafo_lv_q'] = abs(trafo_temp.q_lv) * sigma_trafo + 1.
        trafo_temp['element_type'] = 'trafo'
        trafo_temp['element'] = trafo_temp.index[:]
        # copy the data into the measurement dataframe
        # ---------------------------------------measure: trafo p hv---------------------------------------------------
        trafo_temp['measurement_type'] = 'p'
        trafo_temp['value'] = trafo_temp.p_hv
        trafo_temp['std_dev'] = trafo_temp.stddev_trafo_hv_p
        trafo_temp['side'] = trafo_temp.hv_bus
        trafo_temp = pd.merge(trafo_temp, sv_powerflow, left_on=sc['t_hv'], right_index=True)
        self._copy_to_measurement(trafo_temp)
        trafo_temp = trafo_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo p lv---------------------------------------------------
        trafo_temp['value'] = trafo_temp.p_lv
        trafo_temp['std_dev'] = trafo_temp.stddev_trafo_lv_p
        trafo_temp['side'] = trafo_temp.lv_bus
        trafo_temp = pd.merge(trafo_temp, sv_powerflow, left_on=sc['t_lv'], right_index=True)
        self._copy_to_measurement(trafo_temp)
        trafo_temp = trafo_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo q hv---------------------------------------------------
        trafo_temp['measurement_type'] = 'q'
        trafo_temp['value'] = trafo_temp.q_hv
        trafo_temp['std_dev'] = trafo_temp.stddev_trafo_hv_q
        trafo_temp['side'] = trafo_temp.hv_bus
        trafo_temp = pd.merge(trafo_temp, sv_powerflow, left_on=sc['t_hv'], right_index=True)
        self._copy_to_measurement(trafo_temp)
        trafo_temp = trafo_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo q lv---------------------------------------------------
        trafo_temp['value'] = trafo_temp.q_lv
        trafo_temp['std_dev'] = trafo_temp.stddev_trafo_lv_q
        trafo_temp['side'] = trafo_temp.lv_bus
        trafo_temp = pd.merge(trafo_temp, sv_powerflow, left_on=sc['t_lv'], right_index=True)
        self._copy_to_measurement(trafo_temp)

        # ---------------------------------------measure: trafo3w---------------------------------------------------
        sigma_trafo3w = 0.03
        trafo3w_temp = self.net.trafo3w[
            ['name', 'hv_bus', 'mv_bus', 'lv_bus', sc['t_hv'], sc['t_mv'], sc['t_lv']]].copy()
        trafo3w_temp['p_hv'] = \
            pd.merge(trafo3w_temp[sc['t_hv']], sv_powerflow['p'], left_on=sc['t_hv'], right_index=True)['p']
        trafo3w_temp['p_mv'] = \
            pd.merge(trafo3w_temp[sc['t_mv']], sv_powerflow['p'], left_on=sc['t_mv'], right_index=True)['p']
        trafo3w_temp['p_lv'] = \
            pd.merge(trafo3w_temp[sc['t_lv']], sv_powerflow['p'], left_on=sc['t_lv'], right_index=True)['p']
        trafo3w_temp['q_hv'] = \
            pd.merge(trafo3w_temp[sc['t_hv']], sv_powerflow['q'], left_on=sc['t_hv'], right_index=True)['q']
        trafo3w_temp['q_mv'] = \
            pd.merge(trafo3w_temp[sc['t_mv']], sv_powerflow['q'], left_on=sc['t_mv'], right_index=True)['q']
        trafo3w_temp['q_lv'] = \
            pd.merge(trafo3w_temp[sc['t_lv']], sv_powerflow['q'], left_on=sc['t_lv'], right_index=True)['q']

        trafo3w_temp = trafo3w_temp.dropna(subset=['p_hv', 'p_mv', 'p_lv', 'q_hv', 'q_mv', 'q_lv'], thresh=6)

        trafo3w_temp['stddev_trafo_hv_p'] = abs(trafo3w_temp.p_hv) * sigma_trafo3w + 1.
        trafo3w_temp['stddev_trafo_mv_p'] = abs(trafo3w_temp.p_mv) * sigma_trafo3w + 1.
        trafo3w_temp['stddev_trafo_lv_p'] = abs(trafo3w_temp.p_lv) * sigma_trafo3w + 1.
        trafo3w_temp['stddev_trafo_hv_q'] = abs(trafo3w_temp.q_hv) * sigma_trafo3w + 1.
        trafo3w_temp['stddev_trafo_mv_q'] = abs(trafo3w_temp.q_mv) * sigma_trafo3w + 1.
        trafo3w_temp['stddev_trafo_lv_q'] = abs(trafo3w_temp.q_lv) * sigma_trafo3w + 1.
        trafo3w_temp['element_type'] = 'trafo3w'
        trafo3w_temp['element'] = trafo3w_temp.index[:]
        # copy the data into the measurement dataframe
        # ---------------------------------------measure: trafo3w p hv---------------------------------------------
        trafo3w_temp['measurement_type'] = 'p'
        trafo3w_temp['value'] = trafo3w_temp.p_hv
        trafo3w_temp['std_dev'] = trafo3w_temp.stddev_trafo_hv_p
        trafo3w_temp['side'] = trafo3w_temp.hv_bus
        trafo3w_temp = pd.merge(trafo3w_temp, sv_powerflow, left_on=sc['t_hv'], right_index=True)
        self._copy_to_measurement(trafo3w_temp)
        trafo3w_temp = trafo3w_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo3w p mv---------------------------------------------
        trafo3w_temp['value'] = trafo3w_temp.p_mv
        trafo3w_temp['std_dev'] = trafo3w_temp.stddev_trafo_mv_p
        trafo3w_temp['side'] = trafo3w_temp.mv_bus
        trafo3w_temp = pd.merge(trafo3w_temp, sv_powerflow, left_on=sc['t_mv'], right_index=True)
        self._copy_to_measurement(trafo3w_temp)
        trafo3w_temp = trafo3w_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo3w p lv---------------------------------------------
        trafo3w_temp['value'] = trafo3w_temp.p_lv
        trafo3w_temp['std_dev'] = trafo3w_temp.stddev_trafo_lv_p
        trafo3w_temp['side'] = trafo3w_temp.lv_bus
        trafo3w_temp = pd.merge(trafo3w_temp, sv_powerflow, left_on=sc['t_lv'], right_index=True)
        self._copy_to_measurement(trafo3w_temp)
        trafo3w_temp = trafo3w_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo3w q hv---------------------------------------------
        trafo3w_temp['measurement_type'] = 'q'
        trafo3w_temp['value'] = trafo3w_temp.q_hv
        trafo3w_temp['std_dev'] = trafo3w_temp.stddev_trafo_hv_q
        trafo3w_temp['side'] = trafo3w_temp.hv_bus
        trafo3w_temp = pd.merge(trafo3w_temp, sv_powerflow, left_on=sc['t_hv'], right_index=True)
        self._copy_to_measurement(trafo3w_temp)
        trafo3w_temp = trafo3w_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo3w q mv---------------------------------------------
        trafo3w_temp['value'] = trafo3w_temp.q_mv
        trafo3w_temp['std_dev'] = trafo3w_temp.stddev_trafo_mv_q
        trafo3w_temp['side'] = trafo3w_temp.mv_bus
        trafo3w_temp = pd.merge(trafo3w_temp, sv_powerflow, left_on=sc['t_mv'], right_index=True)
        self._copy_to_measurement(trafo3w_temp)
        trafo3w_temp = trafo3w_temp.drop(columns=sv_powerflow.columns)
        # ---------------------------------------measure: trafo3w q lv---------------------------------------------
        trafo3w_temp['value'] = trafo3w_temp.q_lv
        trafo3w_temp['std_dev'] = trafo3w_temp.stddev_trafo_lv_q
        trafo3w_temp['side'] = trafo3w_temp.lv_bus
        trafo3w_temp = pd.merge(trafo3w_temp, sv_powerflow, left_on=sc['t_lv'], right_index=True)
        self._copy_to_measurement(trafo3w_temp)

        # remove NaN values
        self.net.measurement = self.net.measurement.dropna(subset=['value'])
        # set the element from float to default uint32
        self._set_measurement_element_datatype()

        self.logger.info("Needed time for creating the measurements: %ss" % (time.time() - time_start))
